package com.ringlord;

import com.ringlord.convert.Value;


/**
 * A one- or two-ended range, where each end defines a value and a comparison.
 * For example, a one-ended range could define "&lt;= 20", whereas a two-ended
 * range might define "&lt;= 20 OR &gt; 40".
 * 
 * @author K Udo Schuermann
 * @param <T>
 *          A {@link Comparable} type, such as {@link String}, {@link Long},
 *          {@link java.util.Date}, etc.
 */
public class ComparableRange<T extends Comparable<T>>
{
  /**
   * Constructs a ComparableRange.
   * 
   * @param firstValue
   *          The {@link Value} for the first end of the Range. It must not be
   *          null.
   */
  public ComparableRange( final T firstValue,
	                  final T secondValue )
  {
    super();
    if( firstValue == null )
      {
	throw new IllegalArgumentException( "The first value of the range cannot be null" );
      }
    this.firstValue = firstValue;
    this.secondValue = secondValue;
  }


  public T getFirstValue()
  {
    return firstValue;
  }


  public T getSecondValue()
  {
    return secondValue;
  }


  public void setFirstComparator( final Comparator comparator )
  {
    this.firstComparator = comparator;
  }


  public void setSecondComparator( final Comparator comparator )
  {
    this.secondComparator = comparator;
  }


  /**
   * Determine if the given value matches the defined range.
   * 
   * @param value
   *          The value to be compared to the defined range. It must not be
   *          null.
   * @return 'true' if the value matches the defined range, 'false' if it does
   *         not.
   */
  public boolean isMatch( final T value )
  {
    final boolean result;

    if( secondValue == null )
      {
	result = matches( value,
	                  firstComparator,
	                  firstValue );
      }
    else
      {
	final boolean isInRangeL = matches( value,
	                                    firstComparator,
	                                    firstValue );
	final boolean isInRangeH = matches( value,
	                                    secondComparator,
	                                    secondValue );

	if( isDisjointComparison( firstComparator,
	                          secondComparator ) )
	  {
	    result = (isInRangeL || isInRangeH);
	  }
	else
	  {
	    result = (isInRangeL && isInRangeH);
	  }
      }
    return result;
  }


  /**
   * Determine whether the comparisons of the <em>two ends of the range</em>
   * form a disjointed set so that an OR comparison is an appropriate
   * comparison, otherwise an AND operation will be performed.
   * 
   * @param firstComparator
   *          The first {@link Comparator}. It must not be null.
   * 
   * @param secondComparator
   *          The second {@link Comparator}. It must not be null.
   * 
   * @return 'true' if an OR operation applies, rather than AND.
   * 
   * @throws IllegalStateException
   *           The range is a one-ended range, meaning that it does not define a
   *           second end. You must call {@link #setSecondValue(Value)} with a
   *           non-null parameter to define a two-sided range.
   */
  public static boolean isDisjointComparison( final Comparator firstComparator,
	                                      final Comparator secondComparator )
  {
    if( (firstComparator == null) || (secondComparator == null) )
      {
	throw new IllegalArgumentException( "Neither firstComparator nor secondComparator can be null" );
      }

    final boolean result;

    if( (firstComparator == Comparator.EQ) && (secondComparator == Comparator.EQ) )
      {
	// Both are EQ (either #1 or #2)
	result = true; // OR
      }
    else if( (firstComparator == Comparator.NE) && (secondComparator == Comparator.NE) )
      {
	// Not X and not Y
	result = false; // AND
      }
    else if( ((firstComparator == Comparator.LT) || (firstComparator == Comparator.LE)) &&
	     ((secondComparator == Comparator.GE) || (secondComparator == Comparator.GT)) )
      {
	// First says < second is > (or <= and >=)
	result = true; // AND
      }
    else if( ((firstComparator == Comparator.NE) || (secondComparator == Comparator.NE)) &&
	     ((firstComparator != Comparator.EQ) && (secondComparator != Comparator.EQ)) &&
	     (firstComparator != secondComparator) )
      {
	// One (but not both) is NE but the other cannot be NE or EQ
	result = false;
      }
    else
      {
	result = false; // OR
      }

    return result;
  }


  /**
   * Determine whether the given two comparisons ("LT", "LE", "EQ", "GE", "GT")
   * form an intersecting set so that an AND comparison is appropriate.
   * 
   * @param c1
   *          The left-side comparison
   * @param c2
   *          The right-side comparison
   * @return 'true' if an AND operation applies, rather than OR.
   */
  public static boolean isIntersectingComparison( final Comparator firstComparator,
	                                          final Comparator secondComparator )
  {
    final boolean result;
    if( ((firstComparator == Comparator.GT) || (firstComparator == Comparator.GE)) &&
	((secondComparator == Comparator.LE) || (secondComparator == Comparator.LT)) )
      {
	result = false;
      }
    else if( ((firstComparator == Comparator.NE) || (secondComparator == Comparator.NE)) &&
	     ((firstComparator != Comparator.EQ) && (secondComparator != Comparator.EQ)) &&
	     (firstComparator != secondComparator) )
      {
	// One (but not both) is NE but the other cannot be NE or EQ
	result = true; // intersects so long as the NE value is within the range
      }
    else if( (firstComparator == Comparator.NE) && (secondComparator == Comparator.NE) )
      {
	// Not X and not Y
	result = true;
      }
    else
      {
	result = false;
      }
    return result;
  }


  /**
   * Perform the defined comparison of a given value and the "constant" that the
   * caller parsed from the editor. This method is called by {@link #isMatch}.
   * 
   * @param value
   * @param comparison
   * @param constantValue
   * @return
   */
  private boolean matches( final T value,
	                   final Comparator comparator,
	                   final T constantValue )
  {
    final boolean result;

    switch( comparator )
      {
      case LT:
	result = (value.compareTo( constantValue ) < 0);
	break;

      case LE:
	result = (value.compareTo( constantValue ) <= 0);
	break;

      case EQ:
	result = (value.compareTo( constantValue ) == 0);
	break;

      case NE:
	result = (value.compareTo( constantValue ) != 0);
	break;

      case GE:
	result = (value.compareTo( constantValue ) >= 0);
	break;

      case GT:
	result = (value.compareTo( constantValue ) > 0);
	break;

      default:
	result = false;
      }

    return result;
  }


  /**
   * A way in which a value can be compared to one end of a
   * {@link ComparableRange}.
   * 
   * @author K Udo Schuermann
   */
  public enum Comparator
  {
    LT("<"),
    LE("\u2264"), // less-than-or-equal symbol
    EQ("="),
    NE("\u2260"), // not-equal symbol
    GE("\u2265"), // greater-than-or-equal symbol
    GT(">");

    private Comparator( final String visual )
    {
      this.visual = visual;
    }


    public static Comparator fromString( final String text,
	                                 final Comparator defaultValue )
    {
      if( text != null )
	{
	  for( final Comparator c : Comparator.values() )
	    {
	      final String cs = c.toString();
	      if( cs.equalsIgnoreCase( text ) )
		{
		  return c;
		}
	    }
	}
      return defaultValue;
    }


    public String getVisual()
    {
      return visual;
    }

    private final String visual;
  }

  private Comparator firstComparator;
  private final T firstValue;
  private Comparator secondComparator;
  private final T secondValue;
}
