package com.ringlord;

import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import com.ringlord.mime.Base64;


/**
 * <p>
 * A simple-to-use way to persist and retrieve settings for an application. The
 * following demonstrates ways to use this class:
 * </p>
 * 
 * <dl>
 * <dt><code>Settings config = getInstance();</code>
 * <dd>Obtain the the Default Settings. The first time this method is called,
 * the caller is considered the owner and the Settings are constructed for that
 * package.
 * <dt><code>Settings config = getInstance( getClass() );</code>
 * <dd>Obtain the Settings for a specifid class (package). If there are no
 * Default Settings defined, then this call will register the returned Settings
 * as the Default Settings.
 * <dt><code>String foo = config.get( "my-value", (String)null );</code>
 * <dd>Obtain a String setting identified by "my-value", and return null if it
 * has no value.
 * <dt><code>boolean bar = config.get( "light-on", true );</code> Obtain a
 * boolean value identified by "light-on"; if it has no value yet, then return
 * true (otherwise the default is ignored).
 * <dt><code>List&lt;String&gt; names = get( "names", null, "{sep}" );</code>
 * <dd>Obtain a list of Strings identified by "names"; no default is returned if
 * the list is not defined; items in the list are separated with the
 * multi-character sequence "{sep}".
 * </dl>
 * 
 * @author K. Udo Schuermann
 **/
public class Config
{
  public Config( final File file )
  {
    this( file,
	  null );
  }


  public Config( final File file,
	         final String commentLine )
  {
    super();
    this.sections = new LinkedHashMap<>();
    this.file = file;
    this.info = (commentLine == null
	? null
	: commentLine.replace( '\n',
	                       ' ' ).replace( '\r',
	                                      ' ' ));
    load( file );
  }


  /**
   * Creates a sub-component config
   * 
   * @param label
   */
  private Config( final String label )
  {
    super();
    this.sections = null;
    this.file = null;
    this.info = label;
  }


  /**
   * Loads the file when the Config is first constructed.
   * 
   * @param file
   */
  private void load( final File file )
  {
    try( final BufferedReader f = new BufferedReader( new FileReader( file ) ) )
      {
	/*
	 * String s; while( (s = f.readLine()) != null ) { final int len =
	 * s.length(); for( int i=0; i<len; i++ ) {
	 * 
	 * } }
	 */
	final Properties p = new Properties();
	p.load( f );
	for( final Object o : p.keySet() )
	  {
	    final String key = (String)o;
	    prefs.put( key,
		       p.getProperty( key ) );
	  }
      }
    catch( final FileNotFoundException x )
      {
	// It's fine, we will create it
      }
    catch( final IOException x )
      {
	x.printStackTrace();
      }
  }


  private void save( final File file )
  {
    try( final BufferedWriter f = new BufferedWriter( new FileWriter( file ) ) )
      {
	final List<String> keys = new ArrayList<String>();
	keys.addAll( prefs.keySet() );
	Collections.sort( keys );

	f.write( "# Edit only at the risk of losing everyting!" + EOLN );
	if( info != null )
	  {
	    f.write( "# " + info + EOLN );
	  }
	f.write( "# Last written on " + new Date() + EOLN );
	f.write( EOLN );
	for( final String key : keys )
	  {
	    f.write( key + "=" + escape( prefs.get( key ) ) + EOLN );
	  }
	f.write( EOLN );
	f.write( "#eot" );
	f.flush();
      }
    catch( final IOException x )
      {
	x.printStackTrace();
      }
  }


  /**
   * The section name, or null if it is not a section.
   * 
   * @return The name of this sub-config, or null if it is the top-level config.
   */
  public String sectionName()
  {
    return (sections == null
	? info
	: null);
  }


  /**
   * The available section names in the top-level config, or null if this is a
   * section rather than the top-level config.
   * 
   * @return The list of 0+ section names, or null if this is a section.
   */
  public String[] sectionNames()
  {
    if( sections == null )
      {
	return null;
      }
    final String[] result = new String[sections.size()];
    int i = 0;
    for( final Config config : sections.values() )
      {
	result[i++] = config.sectionName();
      }
    return result;
  }


  public Config getSection( final String name )
  {
    return (sections == null
	? null
	: sections.get( name ));
  }


  public void put( final String key,
	           final List<String> value,
	           final String separator )
  {
    if( value == null )
      {
	prefs.remove( key );
      }
    else
      {
	final StringBuilder sb = new StringBuilder();
	boolean isFirst = true;
	if( separator != null )
	  {
	    for( final String s : value )
	      {
		if( isFirst )
		  {
		    isFirst = false;
		  }
		else
		  {
		    sb.append( separator );
		  }
		sb.append( s );
	      }
	  }
	else
	  {
	    for( final String s : value )
	      {
		if( isFirst )
		  {
		    isFirst = false;
		  }
		else
		  {
		    sb.append( "," );
		  }
		final byte[] bytes = Base64.encode( s.getBytes( utf8 ) );
		sb.append( new String( bytes ) );
	      }
	  }
	prefs.put( key,
	           sb.toString() );
      }
    save( file );
  }


  /**
   * Obtain a List of (String) items.
   * 
   * @param key
   *          The key for the list
   * @param defaultValue
   *          A default List if the setting is undefined
   * @param separator
   *          An optional separator: If no separator is defined (null or
   *          zero-length), the items are considered Base64 encoded and
   *          separated with commas, otherwise the separator (which may be
   *          anything, including newline or even multiple characters) is used
   *          to break the value into multiple items.
   * @return
   */
  public List<String> get( final String key,
	                   final List<String> defaultValue,
	                   final String separator )
  {
    final String tmp = prefs.get( key );
    if( tmp == null )
      {
	return defaultValue;
      }
    final List<String> result = new ArrayList<String>();
    int start = 0;
    int pos;
    if( (separator != null) && (separator.length() > 0) )
      {
	final int len = separator.length();
	while( (pos = tmp.indexOf( separator,
	                           start )) > -1 )
	  {
	    result.add( tmp.substring( start,
		                       pos ) );
	    start = pos + len;
	  }
	result.add( tmp.substring( start ) );
      }
    else
      {
	while( (pos = tmp.indexOf( ",",
	                           start )) > -1 )
	  {
	    final byte[] bytes = Base64.decode( tmp.substring( start,
		                                               pos ).getBytes() );
	    result.add( new String( bytes,
		                    utf8 ) );
	    start = pos + 1;
	  }
	final byte[] bytes = Base64.decode( tmp.substring( start ).getBytes() );
	result.add( new String( bytes,
	                        utf8 ) );
      }
    return result;
  }


  // ======================================================================

  public void put( final String key,
	           final String value )
  {
    if( value == null )
      {
	prefs.remove( key );
      }
    else
      {
	prefs.put( key,
	           value );
      }
    save( file );
  }


  public String get( final String key,
	             final String defaultValue )
  {
    final String value = prefs.get( key );
    return (value == null
	? defaultValue
	: value);
  }


  // ======================================================================
  public void put( final String key,
	           final Rectangle rect )
  {
    if( rect == null )
      {
	prefs.remove( key );
      }
    else
      {
	prefs.put( key,
	           rect.x + "," + rect.y + "," + rect.width + "," + rect.height );
      }
    save( file );
  }


  public Rectangle get( final String key,
	                final Rectangle defaultValue )
  {
    final String value = prefs.get( key );
    if( value == null )
      {
	return defaultValue;
      }

    try
      {
	int prev, comma;

	comma = value.indexOf( ',' );
	final int x = Integer.parseInt( value.substring( 0,
	                                                 comma ) );

	prev = comma + 1;
	comma = value.indexOf( ',',
	                       prev );
	final int y = Integer.parseInt( value.substring( prev,
	                                                 comma ) );

	prev = comma + 1;
	comma = value.indexOf( ',',
	                       prev );
	final int w = Integer.parseInt( value.substring( prev,
	                                                 comma ) );

	prev = comma + 1;
	final int h = Integer.parseInt( value.substring( prev ) );

	return new Rectangle( x,
	                      y,
	                      w,
	                      h );
      }
    catch( final Exception x )
      {
	x.printStackTrace();
	return defaultValue;
      }
  }


  // ======================================================================
  public void put( final String key,
	           final Boolean value )
  {
    if( value == null )
      {
	prefs.remove( key );
      }
    else
      {
	prefs.put( key,
	           String.valueOf( value ) );
      }
    save( file );
  }


  public boolean get( final String key,
	              final boolean defaultValue )
  {
    try
      {
	return Boolean.parseBoolean( prefs.get( key ) );
      }
    catch( final NullPointerException x )
      {
	return defaultValue;
      }
  }


  // ======================================================================
  public void put( final String key,
	           final Double value )
  {
    if( value == null )
      {
	prefs.remove( key );
      }
    else
      {
	prefs.put( key,
	           String.valueOf( value ) );
      }
    save( file );
  }


  public double get( final String key,
	             final double defaultValue )
  {
    try
      {
	return Double.parseDouble( prefs.get( key ) );
      }
    catch( NullPointerException |
	   NumberFormatException x )
      {
	return defaultValue;
      }
  }


  // ======================================================================

  public void put( final String key,
	           final Integer value )
  {
    if( value == null )
      {
	prefs.remove( key );
      }
    else
      {
	prefs.put( key,
	           String.valueOf( value ) );
      }
    save( file );
  }


  public int get( final String key,
	          final int defaultValue )
  {
    try
      {
	return Integer.parseInt( prefs.get( key ) );
      }
    catch( NullPointerException |
	   NumberFormatException x )
      {
	return defaultValue;
      }
  }


  // ======================================================================

  private static String escape( final String s )
  {
    final StringBuilder sb = new StringBuilder();
    final int len = s.length();
    for( int i = 0; i < len; i++ )
      {
	final char c = s.charAt( i );
	if( (c == '\\') || (c == '\n') || (c == '\r') )
	  {
	    sb.append( '\\' );
	  }
	sb.append( c );
      }
    return sb.toString();
  }

  private final Map<String,Config> sections;
  private final File file;
  private final String info;
  private final Map<String,String> prefs = new HashMap<String,String>();
  //
  private static final Charset utf8 = Charset.forName( "UTF-8" );
  private static final String EOLN = System.getProperty( "line.separator" );
}
