package com.ringlord;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;


/**
 * Provides a mechanism to invoke an external program with parameters, supply it
 * stdin, and capture stdout and stderr without the risk of blocking and hanging
 * the process.
 * 
 * @author K. Udo Schuermann
 **/
public class Runner
{
  public static void execute( final String... commandline )
    throws IOException
  {
    execute( null,
	     null,
	     null,
	     commandline );
  }


  /**
   * Using input and output streams (any of which are optional and can be null),
   * invoke a given program and return only when the program has terminated.
   **/
  public static void execute( final InputStream stdin,
	                      final OutputStream stdout,
	                      final OutputStream stderr,
	                      final String... commandline )
    throws IOException
  {
    final Process p = Runtime.getRuntime().exec( commandline );

    final Firehose out = new Firehose( p.getInputStream(),
	                               stdout,
	                               "stdout: " + commandline[0],
	                               512 );
    out.start();

    final Firehose err = new Firehose( p.getErrorStream(),
	                               stderr,
	                               "stderr: " + commandline[0],
	                               512 );
    err.start();

    final OutputStream input = p.getOutputStream();
    if( stdin != null )
      {
	final Firehose in = new Firehose( stdin,
	                                  input,
	                                  "stdin: " + commandline[0],
	                                  512 );
	in.start();
	try
	  {
	    in.join();
	  }
	catch( final InterruptedException x )
	  {
	  }
      }
    input.close();

    try
      {
	p.waitFor();
      }
    catch( final InterruptedException x )
      {
      }

    try
      {
	out.join();
      }
    catch( final InterruptedException x )
      {
      }

    try
      {
	err.join();
      }
    catch( final InterruptedException x )
      {
      }
  }


  private Runner()
  {
  }


  public static class Firehose
    extends Thread
  {
    public Firehose( final InputStream in,
	             final OutputStream out )
    {
      this( in,
	    out,
	    "Firehose",
	    4096 );
    }


    public Firehose( final InputStream in,
	             final OutputStream out,
	             final String name,
	             final int bufferSize )
    {
      super( name );
      this.in = in;
      this.out = out;
    }


    @Override
    public void run()
    {
      final byte[] buffer = new byte[512];
      try
	{
	  int inBuffer;
	  if( out == null )
	    {
	      while( (inBuffer = in.read( buffer )) > -1 )
		{
		  // no stream just eat all the input
		}
	    }
	  else
	    {
	      while( (inBuffer = in.read( buffer )) > -1 )
		{
		  out.write( buffer,
			     0,
			     inBuffer );
		}
	      out.flush();
	    }
	}
      catch( final IOException x )
	{
	  x.printStackTrace();
	}
    }
    private final InputStream in;
    private final OutputStream out;
  }
}
