package com.ringlord;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;


/**
 * <p>
 * A Version represents a major/minor/patch/... sequence of numbers followed by
 * an optional character-based suffix. A Version is not mutable.
 * </p>
 * 
 * <p>
 * A version is composed of at least one number (such as Firefox "9") but two
 * are found more traditionally (e.g. Linux Kernel "3.2"). A third digit tends
 * to indicate a patch to a minor release (e.g. LibreOffice "3.4.6") and even
 * lesser patch numbers are supported without limit.
 * </p>
 * 
 * <p>
 * Any non-digit is treated as a separator between version number components;
 * separators are discarded.
 * </p>
 **/
public class Version
  implements
    Iterable<String>,
    Comparable<Version>
{
  public static void main( final String[] args )
  {
    final List<Version> collected = new ArrayList<Version>();
    for( final String s : args )
      {
	final Version v = Version.parse( s );
	collected.add( v );
      }
    java.util.Collections.sort( collected );
    for( final Version v : collected )
      {
	System.err.println( "*** " + v );
      }
  }


  /**
   * Create a Version from an arbitrarily formatted version identifier where
   * unbroken sequences of digits represent a single component of the version,
   * non-digits between such sequences are separators (and discard) and trailing
   * non-digits are kept as a suffix.
   **/
  public static Version parse( final String deweyDecimal )
  {
    final List<Integer> values = new ArrayList<Integer>();

    final int len = deweyDecimal.length();
    int tail = -1;
    int start = -1;
    for( int i = 0; i < len; i++ )
      {
	final char c = deweyDecimal.charAt( i );
	if( "0123456789".indexOf( c ) > -1 )
	  {
	    if( start < 0 )
	      {
		start = i;
	      }
	    continue;
	  }
	if( start > -1 )
	  {
	    values.add( Integer.valueOf( deweyDecimal.substring( start,
		                                                 i ) ) );
	    tail = i;
	    start = -1;
	  }
      }
    if( start > -1 )
      {
	values.add( Integer.valueOf( deweyDecimal.substring( start ) ) );
	tail = -1;
      }
    final int[] components = new int[values.size()];
    int i = 0;
    for( final Integer n : values )
      {
	components[i++] = n;
      }
    if( (tail >= 0) && (tail < len) )
      {
	return new Version( components,
	                    deweyDecimal.substring( tail ) );
      }
    else
      {
	return new Version( components,
	                    null );
      }
  }


  public Version( final int[] components )
  {
    this( components,
	  null );
  }


  /**
   * Create a Version from a pre-processed set of numeric components, and an
   * optional suffix (the suffix may be null).
   **/
  public Version( final int[] components,
	          final String suffix )
  {
    for( final int n : components )
      {
	this.components.add( n );
      }
    this.suffix = suffix;
  }


  // ----------------------------------------------------------------------
  @Override
  public String toString()
  {
    if( str == null )
      {
	final StringBuilder sb = new StringBuilder();
	boolean isFirst = true;
	for( final Integer n : components )
	  {
	    if( isFirst )
	      {
		isFirst = false;
	      }
	    else
	      {
		sb.append( "." );
	      }
	    sb.append( String.valueOf( n ) );
	  }
	if( suffix != null )
	  {
	    sb.append( suffix );
	  }
	str = sb.toString();
      }
    return str;
  }


  // ----------------------------------------------------------------------
  @Override
  public int compareTo( final Version other )
  {
    final int v1Len = components.size();
    final int v2Len = other.components.size();
    final int maxLen = Math.max( v1Len,
	                         v2Len );
    for( int i = 0; i < maxLen; i++ )
      {
	final int n1 = (i < v1Len
	    ? components.get( i )
	    : 0);
	final int n2 = (i < v2Len
	    ? other.components.get( i )
	    : 0);
	if( n1 != n2 )
	  {
	    return (n1 - n2);
	  }
      }
    // The common numeric components are equal
    if( v1Len < v2Len )
      {
	return -1;
      }
    else if( v1Len > v2Len )
      {
	return 1;
      }
    // The numeric components are perfectly matched and equal
    if( suffix != null )
      {
	if( other.suffix != null )
	  {
	    return suffix.compareTo( other.suffix );
	  }
	else
	  {
	    // A version with suffix always follows one without
	    return 1;
	  }
      }
    else if( other.suffix != null )
      {
	// A version without suffix always precedes one with suffix
	return -1;
      }
    return 0; // totally equal versions!
  }


  // ----------------------------------------------------------------------
  @Override
  public Iterator<String> iterator()
  {
    final List<String> tmp = new ArrayList<String>();
    for( final Integer n : components )
      {
	tmp.add( String.valueOf( n ) );
      }
    if( suffix != null )
      {
	tmp.add( suffix );
      }
    return tmp.iterator();
  }
  // ----------------------------------------------------------------------
  private String str;
  private final List<Integer> components = new ArrayList<Integer>();
  private final String suffix;
}
