package com.ringlord.archive;

import java.io.Closeable;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;


/**
 * <p>
 * Bare bones TapeArchive (TAR) Reader/Writer classes, intended only for the
 * most common file packaging uses, but lacking support for most functionality
 * of which the tar format and more specific software (such as gnutar) are
 * capable.
 * </p>
 * 
 * <p>
 * The TapeArchive's Reader/Writer classes are not streams, and cannot be used
 * directly as part of a GZipInputStream/GZipOutputStream. They must exist as
 * uncompressed files on disk.
 * </p>
 * 
 * @author K Udo Schuermann
 */
public final class TapeArchive
{
  /**
   * Exposes the contents of TapeArchive (.tar) files.
   * 
   * @author K. Udo Schuermann
   */
  public static class Reader
    implements
      Closeable,
      Iterable<Entry>
  {
    public Reader( final File file )
      throws FileNotFoundException
    {
      super();
      this.file = file;
      this.input = new RandomAccessFile( file,
	                                 "r" );
    }


    /**
     * <p>
     * Produces the set of Tape Archive {@linkplain Entry file entries}. The
     * first time this method is invoked, the tar file is scanned; additional
     * calls will use a cached result.
     * </p>
     * 
     * <p>
     * Note that the {@link InputStream} from an Entry is backed by the Reader's
     * own input, so {@linkplain #close() closing the Reader} will cause all
     * InputStreams of its Entries to fail additional I/O.
     * </p>
     * 
     * <p>
     * The InputStreams from the Entries support the mark/reset mechanism
     * without limit; the mark is initially at the start of the file and calling
     * reset resets the stream to the very beginning.
     * </p>
     */
    public Iterator<Entry> iterator()
    {
      synchronized( this )
	{
	  if( content == null )
	    {
	      content = new ArrayList<>();
	      try
		{
		  for( ;; )
		    {
		      final Entry tarEntry = nextTarEntry();
		      content.add( tarEntry );
		      input.seek( input.getFilePointer() + tarEntry.size );
		    }
		}
	      catch( IOException x )
		{
		}
	    }
	}
      return content.iterator();
    }


    @Override
    public void close()
      throws IOException
    {
      input.close();
    }


    private Entry nextTarEntry()
      throws IOException
    {
      final long offset = input.getFilePointer();
      final int inBuffer = input.read( HEADER,
	                               0,
	                               512 );
      if( inBuffer < 512 )
	{
	  throw new IOException( "End of file" );
	}

      final StringBuilder name = new StringBuilder();
      for( int i = 0; i < 100; i++ )
	{
	  if( HEADER[i] == 0 )
	    {
	      break;
	    }
	  name.append( (char)HEADER[i] );
	}
      long size;
      if( HEADER[124] == 0 )
	{
	  throw new IOException( "Empty record, end of file" );
	}
      if( HEADER[124] > 127 ) // high order bit set?
	{
	  // A modern base256 value
	  throw new IOException( "Base256 encoded file size not yet supported" );
	}
      else
	{
	  // An old octal only file size
	  final char[] sizeData = {(char)HEADER[124],
		                   (char)HEADER[125],
		                   (char)HEADER[126],
		                   (char)HEADER[127],
		                   (char)HEADER[128],
		                   (char)HEADER[129],
		                   (char)HEADER[130],
		                   (char)HEADER[131],
		                   (char)HEADER[132],
		                   (char)HEADER[133],
		                   (char)HEADER[134]};
	  final String sizeText = new String( sizeData );
	  size = Long.parseLong( sizeText,
		                 8 );
	  final int padding = 512 - (int)(size % 512);
	  System.out.println( "PADDING: " + padding );
	  if( padding < 512 )
	    {
	      System.out.println( "Skipping " + padding + " padding bytes" );
	      input.seek( input.getFilePointer() + padding );
	    }
	}
      final Entry tarEntry = new Entry( name.toString(),
	                                offset,
	                                size,
	                                this );
      // TODO fill the tarEntry with more data from HEADER
      return tarEntry;
    }

    private List<Entry> content;
    //
    private final File file;
    private final RandomAccessFile input;
    private final byte[] HEADER = new byte[512];
  }


  /**
   * Provides the means to create UStar Tape Archive (.tar) files. The Writer is
   * not thread safe.
   * 
   * @author K Udo Schuermann
   */
  public static class Writer
    implements
      Closeable
  {
    public Writer( final File file )
      throws FileNotFoundException
    {
      super();
      this.file = file;
      this.output = new RandomAccessFile( file,
	                                  "rw" );
    }


    /**
     * Add a new TarEntry to the archive, most of whose information is managed
     * and updated for you. If you wish to override uid, gid, file permissions,
     * 
     * @param inArchiveName
     *          To prevent tarbombs and security issues, the name must have at
     *          least one non-leading directory separator ('/'; Microsoft-style
     *          backward slashes will be converted to normal directory
     *          separators), and must not contain any '..' (parent) directory
     *          references.
     * @throws IOException
     */
    public Entry addEntry( final String inArchiveName )
      throws IOException
    {
      closeEntry(); // close it if necessary

      // Verify that the name is not an obvious tarbomb or security risk
      String sanitizedName = inArchiveName.replace( '\\',
	                                            '/' );
      final int firstSlashIndex = sanitizedName.indexOf( "/" );
      if( firstSlashIndex < 1 ) // not there (-1) or leading (0)
	{
	  if( firstSlashIndex < 0 )
	    {
	      throw new IllegalArgumentException( "Name must specify at least one subdirectory level" );
	    }
	  else
	    {
	      throw new IllegalArgumentException( "Name must not begin at the root" );
	    }
	}
      if( sanitizedName.startsWith( "../" ) || sanitizedName.indexOf( "/../" ) > -1 )
	{
	  throw new IllegalArgumentException( "Name cannot contain '..' (parent) directory references" );
	}

      // create the entry
      currentEntry = new Entry( sanitizedName,
	                        output.getFilePointer(),
	                        0,
	                        null );

      writeHeaderRecord( currentEntry ); // afterwards, ready to write data

      return currentEntry;
    }


    public void write( final byte[] buffer )
      throws IOException
    {
      write( buffer,
	     0,
	     buffer.length );
    }


    public void write( final byte[] buffer,
	               int start,
	               int length )
      throws IOException
    {
      if( currentEntry == null )
	{
	  throw new IllegalStateException( "Must first 'add' an entry to the archive" );
	}
      output.write( buffer,
	            start,
	            length );
    }


    @Override
    public void close()
      throws IOException
    {
      closeEntry();
      output.close();
    }


    private void closeEntry()
      throws IOException
    {
      if( currentEntry != null )
	{
	  long currentFilePos = output.getFilePointer();
	  currentEntry.size = (currentFilePos - (currentEntry.offset + 512));

	  final int padding = 512 - (int)(currentEntry.size % 512);
	  if( padding < 512 )
	    {
	      Arrays.fill( HEADER,
		           0,
		           padding,
		           (byte)0 );
	      output.write( HEADER,
		            0,
		            padding );
	      currentFilePos += padding;
	    }

	  // TODO update currentEntry.checksum (set checksum fields to all 0x20,
	  // checksum, replace them)

	  output.seek( currentEntry.offset );
	  try
	    {
	      writeHeaderRecord( currentEntry );
	    }
	  finally
	    {
	      output.seek( currentFilePos );
	    }

	  currentEntry = null;
	}
    }


    /**
     * The caller must position the file pointer (if necessary) at the correct
     * position in the output, and re-position it (if necessary) where needed
     * afterwards.
     * 
     * @param curEntry
     *          The entry whose data is to be written.
     * @throws IOException
     */
    private void writeHeaderRecord( final Entry curEntry )
      throws IOException
    {
      Arrays.fill( HEADER,
	           0,
	           HEADER.length - 1,
	           (byte)0 );

      // Only supports "short" filenames for now
      final int nameLen = curEntry.name.length();
      if( nameLen < 100 )
	{
	  for( int i = 0; i < nameLen; i++ )
	    {
	      HEADER[i] = (byte)curEntry.name.charAt( i );
	    }
	}
      else
	{
	  System.err.println( "NAME TOO LONG" );
	  HEADER[0] = (byte)'X';
	}

      // Fake the filemode to be 0700
      HEADER[100] = '0';
      HEADER[101] = '0';
      HEADER[102] = '0';
      HEADER[103] = '0';
      HEADER[104] = '7';
      HEADER[105] = '0';
      HEADER[106] = '0';

      // Store the filesize (only supports octal size for now)
      final byte[] sizeBytes = String.format( "%011o",
	                                      curEntry.size ).getBytes();
      for( int i = 0; i < 11; i++ )
	{
	  HEADER[124 + i] = sizeBytes[i];
	}

      final byte[] dateBytes = String.format( "%011o",
	                                      curEntry.mtime / 1000 ).getBytes();
      for( int i = 0; i < 11; i++ )
	{
	  HEADER[136 + i] = dateBytes[i];
	}

      HEADER[257] = (byte)'u';
      HEADER[258] = (byte)'s';
      HEADER[259] = (byte)'t';
      HEADER[260] = (byte)'a';
      HEADER[261] = (byte)'r';
      HEADER[262] = (byte)' ';
      HEADER[263] = (byte)' ';

      HEADER[265] = (byte)'n';
      HEADER[266] = (byte)'o';
      HEADER[267] = (byte)'b';
      HEADER[268] = (byte)'o';
      HEADER[269] = (byte)'d';
      HEADER[270] = (byte)'y';

      HEADER[297] = (byte)'n';
      HEADER[298] = (byte)'o';
      HEADER[299] = (byte)'b';
      HEADER[300] = (byte)'o';
      HEADER[301] = (byte)'d';
      HEADER[302] = (byte)'y';

      HEADER[148] = (byte)' ';
      HEADER[149] = (byte)' ';
      HEADER[150] = (byte)' ';
      HEADER[151] = (byte)' ';
      HEADER[152] = (byte)' ';
      HEADER[153] = (byte)' ';
      HEADER[154] = (byte)' ';
      HEADER[155] = (byte)' ';
      int checksum = 0;
      for( int i = 0; i < 512; i++ )
	{
	  checksum += HEADER[i];
	}
      final byte[] cksum = String.format( "%06o",
	                                  checksum ).getBytes();
      for( int i = 0; i < 6; i++ )
	{
	  HEADER[148 + i] = cksum[i];
	}
      HEADER[154] = 0;

      output.write( HEADER,
	            0,
	            512 );
    }
    private Entry currentEntry;
    private final RandomAccessFile output;
    private final File file;
    private static final byte[] HEADER = new byte[512];
  }


  private static class TarInputStream
    extends InputStream
  {
    private TarInputStream( final RandomAccessFile input,
	                    final Entry entry )
    {
      super();
      this.input = input;
      this.offset = entry.offset + 512;
      this.size = entry.size;
    }


    @Override
    public int read()
      throws IOException
    {
      synchronized( input )
	{
	  final long pos = input.getFilePointer();
	  try
	    {
	      input.seek( offset + consumed );
	      return input.read();
	    }
	  finally
	    {
	      input.seek( pos );
	    }
	}
    }


    @Override
    public int read( final byte[] buffer,
	             final int start,
	             final int length )
      throws IOException
    {
      synchronized( input )
	{
	  if( consumed >= size )
	    {
	      return -1; // indicate EOF
	    }
	  final long pos = input.getFilePointer();
	  try
	    {
	      input.seek( offset + consumed );
	      final int avail = (int)(size - consumed);
	      final int obtained = input.read( buffer,
		                               start,
		                               Math.min( avail,
		                                         length ) );
	      consumed += obtained;
	      return obtained;
	    }
	  finally
	    {
	      input.seek( pos );
	    }
	}
    }


    @Override
    public int available()
    {
      return (int)(size - consumed);
    }


    @Override
    public boolean markSupported()
    {
      return true;
    }


    @Override
    public void mark( final int readLimit )
    {
      // Note: We ignore readLimit as we are backed by a RandomAccessFile and
      // need not limit anything at all.
      this.mark = consumed;
    }


    @Override
    public void reset()
    {
      this.consumed = mark;
    }

    private long mark;
    private long consumed;
    private final long size;
    private final long offset;
    private final RandomAccessFile input;
  }


  public static class Entry
  {
    /**
     * Only Reader/Writer can create instances of the TarEntry
     * 
     * @param name
     * @param offset
     * @param size
     */
    private Entry( final String name,
	           final long offset,
	           final long size,
	           final Reader reader )
    {
      super();
      this.name = name;
      this.offset = offset;
      this.size = size;
      this.reader = reader;
    }


    public Entry setName( final String name )
    {
      this.name = name;
      return this;
    }


    public String getName()
    {
      return name;
    }


    public Entry setUID( final String uid )
    {
      this.uid = uid;
      return this;
    }


    public String getUID()
    {
      return uid;
    }


    public Entry setGID( final String gid )
    {
      this.gid = gid;
      return this;
    }


    public String getGID()
    {
      return gid;
    }


    public Entry setUname( final String uname )
    {
      this.uname = uname;
      return this;
    }


    public String getUname()
    {
      return uname;
    }


    public Entry setGname( final String gname )
    {
      this.gname = gname;
      return this;
    }


    public String getGname()
    {
      return gname;
    }


    public Entry setMTime( final long mtime )
    {
      this.mtime = mtime;
      return this;
    }


    public long getMTime()
    {
      return mtime;
    }


    public Entry setType( final char type )
    {
      this.type = type;
      return this;
    }


    public char getType()
    {
      return type;
    }


    /**
     * <p>
     * Obtains an {@link InputStream} backed by the tar archive file itself,
     * which is why the tar archive must not be closed. Multiple streams are
     * supported in parallel through synchronization and appropriate seeking.
     * </p>
     * 
     * <p>
     * The returned InputStream supports {@link InputStream#mark(int)} and
     * {@link InputStream#reset()} without limit. The default mark is at the
     * beginning of the stream, so you may repeatedly reset() the stream to
     * return to the very beginning of it.
     * </p>
     * 
     * @return
     */
    public InputStream getInputStream()
    {
      if( reader != null )
	{
	  return new TarInputStream( reader.input,
		                     this );
	}
      else
	{
	  throw new IllegalStateException( "Must use a TarArchive.Reader (not .Writer) to read file contents" );
	}
    }

    private long offset;
    private long size;
    private Reader reader;
    //
    private String name;
    private String uid;
    private String gid;
    private String uname;
    private String gname;
    private char type;
    //
    private long mtime;
    private String chksum;
    private char linkflag;
    private String linkName;
    private String magic;
    private String devmajor;
    private String devminor;
  }


  /**
   * Do not allow instances of TapeArchive, only {@link TapeArchive.Reader} and
   * {@link TapeArchive.Writer}.
   */
  private TapeArchive()
  {
  }
}
