package com.ringlord.convert;

import java.util.HashMap;
import java.util.Map;


/**
 * <p>
 * Defines the means to convert from one unit to another. Example code:
 * </p>
 * 
 * <pre>
 * // Convert 8.35 KILOGRAMS into POUNDS:
 * double pounds = Unit.KILOGRAMS.convert( 8.35,
 *                                         Unit.POUNDS );
 * 
 * // Convert 5280 FEET into METERS:
 * double meters = Unit.FEET.convert( 8280,
 *                                    Unit.METERS );
 * 
 * // Convert 4920°C into °F:
 * double fahrenheit = Unit.CELSIUS.convert( 4920,
 *                                           Unit.FAHRENHEIT );
 * </pre>
 * 
 * @author K Udo Schuermann
 */
public class Value
{
  // Sun measurements
  public static final double SOLAR_RADIUS_KM = 6.96342e5d;
  public static final double SOLAR_MASS_KG = 1.9891e30d;
  public static final double SOLAR_LUMINOSITY_W = 3.846e26d;
  // Earth measurements
  public static final double EARTH_RADIUS_KM = 6378.1d;
  public static final double EARTH_MASS_KG = 5.97219e24d;
  // Used to measure distances of astronomical scales
  public static final double ASTRONOMICAL_UNIT_KM = 149597870.7d;
  public static final double LIGHTYEAR_KM = 9460730472580.8;
  public static final double PARSEC_KM = 30856775814913.7;
  // Old British Imperial standards, still used by one or two countries in the
  // world
  public static final double MILE_KM = 1.609344d;
  public static final double FOOT_KM = 0.0003048d;
  public static final double YARD_KM = 0.0009144d;
  //
  public static final double YEAR_DAYS_JULIAN = 365.256363004d;
  public static final double DAY_HOURS_SIDERAL = 0.99726968d;


  public static void main( final String[] args )
  {
    showConversion( 1.5,
	            Unit.KILOGRAMS,
	            Unit.POUNDS );
    showConversion( 1.6,
	            Unit.KILOMETERS,
	            Unit.FEET );
    showConversion( 5280,
	            Unit.FEET,
	            Unit.METERS );
    showConversion( 1,
	            Unit.PARSEC,
	            Unit.LIGHTYEARS );
    showConversion( 100,
	            Unit.CELSIUS,
	            Unit.FAHRENHEIT );
    showConversion( 212,
	            Unit.FAHRENHEIT,
	            Unit.KELVIN );
    showConversion( 373.15,
	            Unit.KELVIN,
	            Unit.CELSIUS );
    showConversion( 1,
	            Unit.SOLAR_RADIUS,
	            Unit.EARTH_RADIUS );
    showConversion( 1,
	            Unit.ASTRONOMICAL_UNITS,
	            Unit.SOLAR_RADIUS );
    showConversion( 1,
	            Unit.ASTRONOMICAL_UNITS,
	            Unit.KILOMETERS );
    showConversion( 1,
	            Unit.LIGHTYEARS,
	            Unit.KILOMETERS );
    showConversion( 1,
	            Unit.LIGHTYEARS,
	            Unit.ASTRONOMICAL_UNITS );

    // Parse text
    System.err.println( "4850°F = " + Value.parse( "4850°F" ) + " (parsed)" );
    System.err.println( "0°C    = " + Value.parse( "0°C" ).convertTo( Unit.FAHRENHEIT ) + Unit.FAHRENHEIT.suffix() );
  }


  private static void showConversion( final double value,
	                              final Unit from,
	                              final Unit to )
  {
    System.err.println( value + from.suffix() + " = " + from.convert( value,
	                                                              to ) + to.suffix() );
  }


  public Value( final double value,
	        final Unit unit )
  {
    super();
    this.value = value;
    this.unit = unit;
  }


  /**
   * @see #parse(String,Type)
   */
  public static Value parse( final String s )
  {
    return parse( s,
	          null );
  }


  /**
   * Parses a string with suffix into a Value. Example: "1.55 km" is the
   * equivalent of <code>new&nbsp;Value(1.55,KILOMETERS)</code>
   * 
   * @param s
   *          A value such as "1.55km" or "2950°K" or "4.33 AU": Spaces between
   *          value and suffix are optional. The suffix may be any mix of upper
   *          or lower case.
   * @param type
   *          An optional {@link Type} that limits the available suffixes. This
   *          is primarily of use with units used in stronomical circles, such
   *          as "sol" and "earth" which are in use to represent both mass and
   *          length (radius) suffixes.
   * @return A Value representing the string-equivalence.
   * @throws IllegalArgumentException
   *           The unit suffix is missing or not recognized
   * @throws NumberFormatException
   *           The numeric portion cannot be parsed
   */
  public static Value parse( final String s,
	                     final Type type )
  {
    int lastNonNumberIndex = s.length() - 1;
    while( (lastNonNumberIndex >= 0) && !Character.isDigit( s.charAt( lastNonNumberIndex ) ) )
      {
	lastNonNumberIndex--;
      }
    final double value = Double.parseDouble( s.substring( 0,
	                                                  lastNonNumberIndex + 1 ).trim() );
    final String suffix = s.substring( lastNonNumberIndex + 1 );
    final Unit unit = Unit.bySuffix( suffix.trim(),
	                             type );
    if( unit != null )
      {
	return new Value( value,
	                  unit );
      }
    throw new IllegalArgumentException( "Unknown suffix: '" + suffix + "'" );
  }


  public static double convert( final double value,
	                        final Unit fromUnit,
	                        final Unit toUnit )
  {
    return fromUnit.convert( value,
	                     toUnit );
  }


  public double convertTo( final Unit toUnit )
  {
    return unit.convert( value,
	                 toUnit );
  }


  public double doubleValue()
  {
    return value;
  }


  public Unit unit()
  {
    return unit;
  }


  /**
   * A string representation of the Value. Note that values that cannot be
   * accurately represented in binary floating point will demonstrate rounding
   * errors. A way to avoid this is to use a slightly shorter format for output
   * using, for example, {@link #toString(String)}.
   * 
   * @see #toString(String)
   */
  @Override
  public String toString()
  {
    return value + unit.suffixes[0];
  }


  /**
   * Controlled formatted output.
   * 
   * @param format
   *          The format containing both numeric form and a string placeholder
   *          for the unit suffix. Example: "%1.6d%s" produces something like
   *          "385.249854m" for a value 385.24985398483... meters
   * @return The formatted value.
   * @see #toString()
   */
  public String toString( final String format )
  {
    return String.format( format,
	                  value,
	                  unit.suffixes[0] );
  }


  public enum Type
  {
    MASS,
    LENGTH,
    TIME,
    TEMPERATURE,
    PRESSURE,
    LUMINOSITY;
  }


  public enum Unit
  {
    // Factors convert to Kilograms
    SOLAR_MASS(Type.MASS, SOLAR_MASS_KG, "M𐌏", "sol"),
    EARTH_MASS(Type.MASS, EARTH_MASS_KG, "M𐌈", "earth"),
    KILOGRAMS(Type.MASS, 1.0d, "kg"),
    GRAMS(Type.MASS, 0.001d, "g", "grams"),
    POUNDS(Type.MASS, 0.454d, "lbs", "lb"),

    // Factors convert to Kilometers
    KILOMETERS(Type.LENGTH, 1.0d, "km"),
    METERS(Type.LENGTH, 0.001d, "m"),
    MILES(Type.LENGTH, MILE_KM, "mi"),
    YARDS(Type.LENGTH, YARD_KM, "yd"),
    FEET(Type.LENGTH, FOOT_KM, "ft"),
    ASTRONOMICAL_UNITS(Type.LENGTH, ASTRONOMICAL_UNIT_KM, "AU", "au"),
    SOLAR_RADIUS(Type.LENGTH, SOLAR_RADIUS_KM, "R𐌏", "sol"),
    EARTH_RADIUS(Type.LENGTH, EARTH_RADIUS_KM, "R𐌈", "earth"),
    LIGHTYEARS(Type.LENGTH, LIGHTYEAR_KM, "ly"),
    PARSEC(Type.LENGTH, PARSEC_KM, "pc"),

    // Factors are meaningless, conversion is handled internally
    KELVIN(Type.TEMPERATURE, 0.0d, "°K", "K"),
    CELSIUS(Type.TEMPERATURE, 0.0d, "°C", "C"),
    FAHRENHEIT(Type.TEMPERATURE, 0.0d, "°F", "F"),

    // Factors convert to Watts
    SOLAR_LUMINOSITY(Type.LUMINOSITY, SOLAR_LUMINOSITY_W, "L𐌏", "sol"),
    WATTS(Type.LUMINOSITY, 1.0d, "W", "Watts"),

    // Factors convert to Hours
    YEARS(Type.TIME, YEAR_DAYS_JULIAN * DAY_HOURS_SIDERAL, "yr"),
    DAYS(Type.TIME, DAY_HOURS_SIDERAL, "d"),
    HOURS(Type.TIME, 1.0d, "h"),
    MINUTES(Type.TIME, 1.0d / 60.0d, "m"),
    SECONDS(Type.TIME, 1.0d / 3600.0d, "s"),

    ATMOSPHERES(Type.PRESSURE, 1.0d, "atm"),
    PASCALS(Type.PRESSURE, 101325.0d, "Pa"),
    MILLIBARS(Type.PRESSURE, 1013.25d, "mbar"),
    BARS(Type.PRESSURE, 1.01325d, "bar"),
    MMMERCURY(Type.PRESSURE, 760.0d, "mmHg"),
    INMERCURY(Type.PRESSURE, 29.92d, "inHg"),
    PSI(Type.PRESSURE, 14.696d, "psi");

    private Unit( final Type type,
	          final double multForPrimary,
	          final String... suffixes )
    {
      this.type = type;
      this.multForPrimary = multForPrimary;
      this.suffixes = suffixes;
      for( final String s : suffixes )
	{
	  units.put( s.toUpperCase(),
		     this );
	}
    }


    public Type type()
    {
      return type;
    }


    public static Unit bySuffix( final String suffix )
    {
      return bySuffix( suffix,
	               null );
    }


    public static Unit bySuffix( final String suffix,
	                         final Type type )
    {
      if( type == null )
	{
	  return units.get( suffix.toUpperCase() );
	}
      for( final Unit u : Unit.values() )
	{
	  if( u.type() == type )
	    {
	      for( final String s : u.suffixes )
		{
		  if( s.equalsIgnoreCase( suffix ) )
		    {
		      return u;
		    }
		}
	    }
	}
      return null;
    }


    public String suffix()
    {
      return suffixes[0];
    }


    /**
     * Convert a value in one unit to a standard unit
     * 
     * @param value
     * @param unit
     * @return
     */
    public double convert( final double value,
	                   final Unit toUnit )
    {
      if( toUnit.type != type )
	{
	  throw new IllegalArgumentException( "Cannot convert " + value + " " + type + " to " + toUnit.type );
	}
      if( this == toUnit )
	{
	  // No conversion needed
	  return value;
	}
      if( type == Type.TEMPERATURE )
	{
	  double kelvin;
	  if( this == KELVIN )
	    {
	      kelvin = value;
	    }
	  else if( this == CELSIUS )
	    {
	      kelvin = value + 273.15d;
	    }
	  else if( this == FAHRENHEIT )
	    {
	      kelvin = ((value - 32.0d) * 5.0 / 9.0) + 273.15d;
	    }
	  else
	    {
	      throw new IllegalArgumentException( "Unhandled temperature " + this );
	    }

	  if( toUnit == KELVIN )
	    {
	      return kelvin;
	    }
	  if( toUnit == CELSIUS )
	    {
	      return kelvin - 273.15d;
	    }
	  if( toUnit == FAHRENHEIT )
	    {
	      return ((kelvin - 273.15d) * 9.0d / 5.0d) + 32.0d;
	    }
	  throw new IllegalArgumentException( "Unhandled temperature " + this );
	}
      return (value * multForPrimary) / toUnit.multForPrimary;
    }
    private final Type type;
    private final double multForPrimary;
    private final String[] suffixes;
  }

  private final double value;
  private final Unit unit;
  private static final Map<String,Unit> units = new HashMap<String,Unit>();
}
