package com.ringlord.i18n;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;

import java.nio.charset.Charset;

import java.util.List;
import java.util.Map;
import java.util.HashMap;


/**
 * <p>
 * A simple mechanism to support internationalization.
 * </p>
 * 
 * <p>
 * Although you need not extend the Messages class, it can be useful to do so in
 * order to isolate all of the default text in one place. You'll notice that
 * such a class is rather simple, however; for simple applications (like this
 * example) you can probably skip this part:
 * </p>
 * 
 * <pre>
 * class MyMessages
 *   extends com.ringlord.i18n.Messages
 * {
 *   static Messages getInstance()
 *   {
 *     if( instance == null )
 *       {
 *         instance = Messages.getInstance( "myapp",
 *                                          defaultMessages );
 *       }
 *     return instance;
 *   }
 *   // Initialize your Map:
 *   private static Map&lt;String,String&gt; = new HashMap&lt;String,String&gt;()
 *   {
 *     {
 *       put( "hello", "Hello, how are you?" );
 *       put( "name",  "Please enter your name:" );
 *       put( "bye",   "Good-bye, {0}" );
 *     }
 *   };
 *   private static com.ringlord.i18n.Messages instance;
 * }
 * </pre>
 * 
 * <p>
 * NOTE: You can pass <code>null</code> for the default messages, but then you
 * really MUST have a file "myapp.properties" available, or all of your messages
 * will consist solely of the key identifiers by which you name them (such as
 * "hello", "name", or "bye").
 * </p>
 * 
 * <p>
 * Next, you construct a language-specific .properties file. Let's call it
 * "myapp_fr.properties"; "de" is for Deutschland (Germany). We'll talk about
 * the specific variations in a minute.
 * </p>
 * 
 * <p>
 * Store the following in your myapp_de.properties file:
 * </p>
 * 
 * <pre>
 * hello = Hallo, wie geht es dir?
 * name  = Wie heißt du?
 * bye   = Auf Wiedersehen, {0}
 * </pre>
 * 
 * <p>
 * Now, in your application obtain an instance of the Messages object (either
 * call MyMessages.getInstance() according to the example above, or (after
 * obtaining an instance of Messages directly using some parameters) you can
 * call Messages.getInstance()) and then use the Messages object:
 * </p>
 * 
 * <pre>
 * MyMessages msg = new MyMessages();
 * JLabel hello = new JLabel( msg.get( &quot;hello&quot; ) );
 * JLabel askName = new JLabel( msg.get( &quot;name&quot; ) );
 * JTextField entry = new JTextField( 16 );
 * String goodBye = msg.get( &quot;bye&quot;,
 *                           entry.getText().trim() );
 * </pre>
 * 
 * <p>
 * Calls such as <code>msg.get("hello")</code> should be quite obvious. It
 * retrieves the language-specific string, an falls back on the default language
 * if there is nothing specific defined.
 * </p>
 * 
 * <p>
 * The last line is a little different, it doesn't just retrieve the
 * language-specific text for "bye", but also passes along the text that the
 * user (presumably) entered in the 'entry' field. This is because the "bye"
 * messages included a parameter, "{0}" which is replaced before the message is
 * generated. Different languages may place multiple parameters in different
 * places in a sentence.
 * </p>
 * 
 * <p>
 * Note: Do not try to assemble messages from fragments in your application
 * because that will almost never work properly when you internationalize your
 * application.
 * </p>
 * 
 * <p>
 * Now, what if you wanted to provide a regional variation for your messages,
 * one German file for central Germany, another for those living in Austria?
 * </p>
 * 
 * <p>
 * Here is how it all actually works:
 * </p>
 * 
 * <ol>
 * <li>You provide default messages for your application; these can be null, but
 * then you must really have a default properties file:
 * <li>The default properties file ("myapp.properties") is loaded and anything
 * defined in it will supercede entries already defined. This might be a good
 * way to update messages without recompiling the code. Well, maybe as an
 * emergency patch.
 * <li>The language-specific file (example: "myapp_de.properties") is loaded,
 * where the "de" is defined by your operating system's Locale and explosed to
 * the Java environment through the "user.language" property. Again, anything in
 * that file will supercede entries previously defined.
 * <li>Country-specific variations (example: "myapp_de_AT.properties") will be
 * loaded next. Again, anything defined for that will supercede entries
 * previously defined.
 * </ol>
 * 
 * <p>
 * If your localizations are incomplete, your users will see a mix of the
 * original language and whatever localizations have been completed.
 * </p>
 * 
 * <p>
 * You should focus on your default language first, then on the overall
 * languages, such as German, French; after that, get more specific by going for
 * regional variations (Austrian German, Canadian French, British English vs. US
 * English, etc.) You may find that the variations need not be complete
 * replacements, just small augmentations.
 * </p>
 * 
 * <p>
 * To test your languages, run your application with a "-Duser.language=de"
 * property (define it on the command line) and then regional variations by
 * adding "-Duser.country=GB"
 * </p>
 * 
 * <p>
 * Now, a few notes on the .properties files:
 * </p>
 * 
 * <ol>
 * <li>The files must use UTF-8 encoding
 * <li>A '#' (hash) symbol as the FIRST on a line signifies a comment line that
 * will be ignored.
 * <li>An entry is a key/value pair, where key and value are separated either
 * with an equal symbol ('=') or a colon (':') whichever comes first. The key
 * MUST begin in the first column.
 * <li>Whitespace is always trimmed from the key and the value.
 * <li>A line beginning with a whitespace is appended to the previous line's
 * value.
 * <li>Special sequences \n \r \t are replaced with newline, carriage return,
 * and horizontal tab in the value. Others are not recognized at this time, but
 * you can type any UTF-8 character.
 * <li>The \ symbol can be used to escape any other character, most specifically
 * the '{' which is used to enclose a numbered or named placeholder.
 * <code>test=\{0}</code> produces a literal "{0}" result, rather than
 * substituting the first (0-indexed) parameter.
 * <li>For leading whitespace, escape one with \
 * <li>For trailing whitespace place a \ after the last space (as in
 * "Name?&nbsp;\")&mdash;The space is preserved, the \ is lost because nothing
 * actually follows it (this differs from the way Unix scripts are parsed, where
 * the trailing \ would escape the following newline and connect two lines).
 * </ol>
 * 
 * @author K. Udo Schuermann
 **/
public class Messages
{
  /**
   * Create an instance of Messages; the parameter are applied only if an
   * instance has not yet been created.
   * 
   * @param basename
   *          The primary portion for the .properties files. It must not be
   *          null.
   * 
   * @param messages
   *          The default messages for your application. You may provide
   *          <code>null</code>, but then you must ensure that at least the
   *          default properties file is available, else all of your messages
   *          will result in something like "?(hello)" rather than the
   *          translation of "hello".
   * 
   * @param feedback
   *          A List into which are placed String messages and Exception objects
   *          that are generated or encountered as the various .properties files
   *          are loaded. This may be null if you are not interested, but can be
   *          useful to determine what files are loaded, or what problems are
   *          encountered when they are processed.
   **/
  public static synchronized Messages getInstance( final String basename,
	                                           final Map<String,String> messages,
	                                           final List<Object> feedback )
  {
    if( instance == null )
      {
	Messages.instance = new Messages( basename,
	                                  messages,
	                                  feedback );
      }
    return instance;
  }


  /**
   * Create an instance of Messages; the parameter are applied only if an
   * instance has not yet been created.
   * 
   * @param basename
   *          The primary portion for the .properties files. It must not be
   *          null.
   * 
   * @param messages
   *          The default messages for your application. You may provide
   *          <code>null</code>, but then you must ensure that at least the
   *          default properties file is available, else all of your messages
   *          will result in something like "?(hello)" rather than the
   *          translation of "hello".
   **/
  public static Messages getInstance( final String basename,
	                              final Map<String,String> messages )
  {
    return getInstance( basename,
	                messages,
	                null );
  }


  /**
   * Create an instance of Messages; the parameter are applied only if an
   * instance has not yet been created. This version of getInstance provides no
   * default messages, so you must ensure that the default properties file is
   * available, else all of your messages will result in something like
   * "?(hello)" rather than the translation of "hello".
   * 
   * @param basename
   *          The primary portion for the .properties files. It must not be
   *          null.
   **/
  public static Messages getInstance( final String basename )
  {
    return getInstance( basename,
	                null,
	                null );
  }


  /**
   * Obtain a previously created instance. If no such instance has been created
   * (using any of the other getInstance methods) then an IllegalStateException
   * is thrown.
   **/
  public static Messages getInstance()
  {
    if( instance == null )
      {
	throw new IllegalStateException( "Must create an instance of Messages "
	                                 + "using a getInstance method that takes "
	                                 + "parameters before getting it again!" );
      }
    return instance;
  }


  /**
   * Obtain localized messages text.
   * 
   * @param key
   *          The identifier of the message.
   * 
   * @return The localized message: If the text contained any substitution
   *         variables, they will have been replaced with a <code>?(id)</code>
   *         sequence, where 'id' is the index or name of the substitution
   *         variable.
   * 
   * @see #get(String,Object...)
   * @see #get(String,Map)
   **/
  public String get( final String key )
  {
    String s = messages.get( key );
    if( s == null )
      {
	s = "?(" + key + ")";
      }
    return subst( s,
	          (Map<String,String>)null );
  }


  /**
   * Obtain localized message text whose indexed parameters have been replaced.
   * 
   * @param key
   *          The identifier of the message.
   * 
   * @param parameters
   *          The indexed parameters, which must be referenced in the text as
   *          {0}, {1}, {2}, etc.
   * 
   * @return The localized message. Substitution variables that were not
   *         supplied will have been replaced with a <code>?(id)</code>
   *         sequence, where 'id' is the index or name of the substitution
   *         variable.
   * 
   * @see #get(String)
   * @see #get(String,Map)
   **/
  public String get( final String key,
	             final Object... parameters )
  {
    String s = messages.get( key );
    if( s == null )
      {
	s = "?(" + key + ")";
      }

    final Map<String,String> args = new HashMap<String,String>();
    for( int i = 0; i < parameters.length; i++ )
      {
	args.put( String.valueOf( i ),
	          (parameters[i] == null
	              ? null
	              : parameters[i].toString()) );
      }

    return subst( s,
	          args );
  }


  /**
   * Obtain localized message text whose named parameters have been replaced.
   * Although not as easy to use as the one taking an arbitrary number of values
   * for input, the mnemonic parameter naming may help documentation and avoids
   * potential confusion.
   * 
   * @param key
   *          The identifier of the message.
   * 
   * @param parameters
   *          The named parameters, which must be referenced in the text as
   *          {...} where '...' is any identifying text such as a name.
   * 
   * @return The localized message. Substitution variables that were not
   *         supplied will have been replaced with a <code>?(id)</code>
   *         sequence, where 'id' is the index or name of the substitution
   *         variable.
   * 
   * @see #get(String)
   * @see #get(String,Object...)
   **/
  public String get( final String key,
	             final Map<String,String> parameters )
  {
    String s = messages.get( key );
    if( s == null )
      {
	s = "3(" + key + ")";
      }

    return subst( s,
	          parameters );
  }


  /**
   * Performs the actual substitution.
   **/
  private String subst( final String s,
	                final Map<String,String> parameters )
  {
    final StringBuilder sb = new StringBuilder();
    final int sLen = s.length();
    int state = 0;
    int start = 0;
    for( int i = 0; i < sLen; i++ )
      {
	final char c = s.charAt( i );
	switch( state )
	  {
	  case 0:
	    if( c == '\\' )
	      {
		state = 1;
	      }
	    else if( c == '{' )
	      {
		state = 2;
		start = i + 1;
	      }
	    else
	      {
		sb.append( c );
	      }
	    break;

	  case 1:
	    switch( c )
	      {
	      case 'n':
		sb.append( "\n" );
		break;

	      case 'r':
		sb.append( "\r" );
		break;

	      case 't':
		sb.append( "\t" );
		break;

	      default:
		sb.append( c );
	      }
	    state = 0;
	    break;

	  case 2:
	    if( c == '}' )
	      {
		final String key = s.substring( start,
		                                i );
		final String val = (parameters == null
		    ? "?(" + key + ")"
		    : parameters.get( key ));
		if( val != null )
		  {
		    sb.append( val );
		  }
		state = 0;
	      }
	  }
      }
    if( state == 2 )
      {
	sb.append( s.substring( start ) );
      }
    return sb.toString();
  }


  /**
   * Loads the file, first exactly as named, then from the "etc/" subdirectory.
   * 
   * @return Returns 'true' if at least one message was loaded.
   **/
  private boolean load( final ClassLoader cl,
	                final List<Object> feedback,
	                final String filename )
  {
    InputStream f;
    if( (f = cl.getResourceAsStream( filename )) == null )
      {
	if( (f = cl.getResourceAsStream( "etc/" + filename )) == null )
	  {
	    if( feedback != null )
	      {
		feedback
		    .add( "com.ringlord.Messages: Found neither " + "'" + filename + "' nor 'etc/" + filename + "'" );
	      }
	    return false;
	  }
	else if( feedback != null )
	  {
	    feedback.add( "com.ringlord.Messages: Loading 'etc/" + filename + "'" );
	  }
      }
    else if( feedback != null )
      {
	feedback.add( "com.ringlord.Messages: Loading '" + filename + "'" );
      }
    final BufferedReader r = new BufferedReader( new InputStreamReader( f,
	                                                                utf8 ) );

    boolean loaded = false;
    try
      {
	final StringBuilder val = new StringBuilder();
	String key = null;
	String s;
	while( (s = r.readLine()) != null )
	  {
	    if( !s.startsWith( "#" ) && (s.length() > 0) )
	      {
		if( Character.isWhitespace( s.charAt( 0 ) ) )
		  {
		    val.append( s.trim() );
		  }
		else
		  {
		    if( key != null )
		      {
			messages.put( key,
			              val.toString() );
			key = null;
			val.setLength( 0 );
			loaded = true;
		      }

		    final int firstEqual = s.indexOf( "=" );
		    final int firstColon = s.indexOf( ":" );
		    final int sep;
		    if( firstEqual == -1 )
		      {
			if( firstColon == -1 )
			  {
			    // Is this supposed to be a continuation line?
			    continue;
			  }
			else
			  {
			    sep = firstColon;
			  }
		      }
		    else
		      {
			sep = firstEqual;
		      }

		    key = s.substring( 0,
			               sep ).trim();
		    val.append( s.substring( sep + 1 ).trim() );
		  }
	      }
	  }
	if( key != null )
	  {
	    messages.put( key,
		          val.toString() );
	    loaded = true;
	  }
      }
    catch( Exception x )
      {
	feedback.add( x );
      }
    finally
      {
	try
	  {
	    r.close();
	  }
	catch( IOException x )
	  {
	  }
      }

    return loaded;
  }


  /**
   * Creates the instance
   **/
  private Messages( final String basename,
	            final Map<String,String> messages,
	            final List<Object> feedback )
  {
    super();
    if( basename == null )
      {
	throw new IllegalArgumentException( "The basename for Messages cannot be null!" );
      }
    this.messages = (messages == null
	? new HashMap<String,String>()
	: messages);

    final String country = System.getProperty( "user.country" );
    final String language = System.getProperty( "user.language" );

    final ClassLoader cl = getClass().getClassLoader();
    boolean loaded = false;
    loaded |= load( cl,
	            feedback,
	            basename + ".properties" );
    loaded |= load( cl,
	            feedback,
	            basename + "_" + language + ".properties" );
    loaded |= load( cl,
	            feedback,
	            basename + "_" + language + "_" + country + ".properties" );
    if( !loaded )
      {
	System.err.println( "No Messages found: " + basename + " for " + language + "_" + country );
      }
  }


  /**
   * Not used.
   **/
  private Messages()
  {
    super();
    messages = null;
  }

  private final Map<String,String> messages;
  private static Messages instance;
  private static final Charset utf8 = Charset.forName( "UTF-8" );
}
