package com.ringlord.swing;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import javax.swing.DefaultComboBoxModel;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.MutableComboBoxModel;

import com.ringlord.Settings;


/**
 * A file/directory picker, for either opening or saving. Both manual input is
 * supported, and picking a value via {@link JFileChooser}. The internal
 * components are accessible to support additional configuration.
 * 
 * ActionListenerS registered with the {@link #getJTextField()} will receive
 * notification with an {@link ActionEvent#getActionCommand()} value "PICKED"
 * when an entry is picked via the JFileChooser.
 * 
 * @author K Udo Schuermann
 */
public class FilePicker
  extends JPanel
{
  public static final String FILE_PICKED = "FILE_PICKED";


  public FilePicker()
  {
    this( null,
	  true,
	  JFileChooser.FILES_AND_DIRECTORIES,
	  null,
	  null );
  }


  public FilePicker( final boolean isOpenDialog,
	             final int fileSelectionMode,
	             final Settings settings,
	             final String rememberBaseKey )
  {
    this( null,
	  isOpenDialog,
	  fileSelectionMode,
	  settings,
	  rememberBaseKey );
  }


  public FilePicker( final File file,
	             final boolean isOpenDialog,
	             final int fileSelectionMode,
	             final Settings settings,
	             final String rememberBaseKey )
  {
    super( new BorderLayout() );

    setOpaque( false );
    this.settings = settings;
    this.rememberBaseKey = rememberBaseKey;

    // If we have a Settings object as well as a 'rememberBaseKey' then try to
    // obtain a maximum count (defaulting to 10) for the number of values that
    // we are to remember for the filenames. If we don't have sufficient info
    // to retrieve such a value, then we'll use a value of 1 which is a simple
    // JTextField
    final int rememberCount = ((settings != null) && (rememberBaseKey != null)
	? settings.get( rememberBaseKey + "-max",
	                10 )
	: 1);
    if( rememberCount <= 1 )
      {
	final JTextField t = new JTextField();
	t.setText( (file == null
	    ? ""
	    : file.toString()) );
	this.inputField = t;
      }
    else
      {
	final MutableComboBoxModel<String> model = new DefaultComboBoxModel<>();
	final ComboText t = new ComboText( model,
	                                   rememberCount );
	t.setEditable( true );
	final List<String> rememberedFilenames = new ArrayList<>();
	for( int i = 0; i < rememberCount; i++ )
	  {
	    final String s = settings.get( rememberBaseKey + "-file." + i,
		                           (String)null );
	    if( s == null )
	      {
		break;
	      }
	    rememberedFilenames.add( 0,
		                     s );
	  }
	for( final String s : rememberedFilenames )
	  {
	    t.add( s );
	  }
	if( file != null )
	  {
	    t.add( file.toString() );
	  }
	this.inputField = t;
      }
    final Icon folderIcon = Images.getIcon( "/img/folder.png" );
    this.filePicker = (folderIcon == null
	? new JButton( "\u2026" )
	: new JButton( folderIcon ));

    add( BorderLayout.CENTER,
	 inputField );
    add( BorderLayout.EAST,
	 filePicker );

    filePicker.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
	if( fileChooser == null )
	  {
	    final String curEntry = (inputField instanceof JTextField
		? ((JTextField)inputField).getText().toString()
		: ((ComboText)inputField).getText());
	    if( curEntry.length() == 0 )
	      {
		fileChooser = new JFileChooser();
	      }
	    else
	      {
		final File cur = new File( curEntry );
		final File parent = cur.getParentFile();
		if( parent == null )
		  {
		    fileChooser = new JFileChooser();
		  }
		else
		  {
		    fileChooser = new JFileChooser( parent );
		  }
	      }
	    fileChooser.setFileSelectionMode( fileSelectionMode );
	    if( approveButtonText != null )
	      {
		fileChooser.setApproveButtonText( approveButtonText );
	      }
	  }

	final int result;
	if( isOpenDialog )
	  {
	    result = fileChooser.showOpenDialog( filePicker );
	  }
	else
	  {
	    result = fileChooser.showSaveDialog( filePicker );
	  }

	if( result == JFileChooser.APPROVE_OPTION )
	  {
	    if( inputField instanceof JTextField )
	      {
		final JTextField field = (JTextField)inputField;
		final String t = field.getText();
		final File existing = new File( t );
		final File picked = fileChooser.getSelectedFile();
		String pickedValue;
		if( !existing.isDirectory() && picked.isDirectory() )
		  {
		    final Path p = Paths.get( field.getText() );
		    final String namePortion = p.getFileName().toString();
		    final int nameIndex = t.lastIndexOf( namePortion );
		    if( nameIndex > -1 )
		      {
			pickedValue = Paths.get( picked.toString(),
			                         namePortion ).toString();
		      }
		    else
		      {
			pickedValue = picked.toString();
		      }
		  }
		else
		  {
		    pickedValue = picked.toString();
		  }
		field.setText( pickedValue );
	      }
	    else
	      {
		((ComboText)inputField).add( fileChooser.getSelectedFile().toString() );
	      }
	    nofifyActionListeners();
	  }
      }
      private JFileChooser fileChooser;
    } );
  }


  @Override
  public void requestFocus()
  {
    inputField.requestFocus();
  }


  @Override
  public boolean requestFocus( final boolean temporary )
  {
    return inputField.requestFocus( temporary );
  }


  @Override
  public boolean requestFocusInWindow()
  {
    return inputField.requestFocusInWindow();
  }


  @Override
  public boolean requestFocusInWindow( final boolean temporary )
  {
    return inputField.requestFocus( temporary );
  }


  @Override
  public void setEnabled( final boolean isEnabled )
  {
    super.setEnabled( isEnabled );
    inputField.setEnabled( isEnabled );
    filePicker.setEnabled( isEnabled );
  }


  public JTextField getJTextField()
  {
    if( inputField instanceof JTextField )
      {
	return (JTextField)inputField;
      }
    throw new IllegalStateException( "FilePicker was not configured for a 1-line JTextField" );
  }


  public ComboText getComboText()
  {
    if( inputField instanceof ComboText )
      {
	return (ComboText)inputField;
      }
    throw new IllegalStateException( "FilePicker was not configured for a multi-line ComboText" );
  }


  public void saveComboText()
  {
    if( inputField instanceof ComboText )
      {
	final ComboText t = (ComboText)inputField;
	final DefaultComboBoxModel<String> model = (DefaultComboBoxModel<String>)t.getModel();
	final int max = settings.get( rememberBaseKey + "-max",
	                              10 );
	final int last = Math.min( max,
	                           model.getSize() - 1 );
	for( int i = 1; i <= max; i++ )
	  {
	    if( i <= last )
	      {
		settings.put( rememberBaseKey + "-file." + (i - 1),
		              model.getElementAt( i ) );
	      }
	    else
	      {
		settings.put( rememberBaseKey + "-file." + (i - 1),
		              (String)null );
	      }
	  }
	return;
      }
    throw new IllegalStateException( "FilePicker was not configured for a multi-line ComboText" );
  }


  public JButton getJButton()
  {
    return filePicker;
  }


  private void nofifyActionListeners()
  {
    final ActionEvent e = new ActionEvent( inputField,
	                                   ActionEvent.ACTION_PERFORMED,
	                                   FILE_PICKED );
    final ActionListener[] actionListeners;
    if( inputField instanceof JTextField )
      {
	actionListeners = ((JTextField)inputField).getActionListeners();
      }
    else
      {
	actionListeners = ((ComboText)inputField).getActionListeners();
      }
    for( final ActionListener l : actionListeners )
      {
	l.actionPerformed( e );
      }
  }

  private String approveButtonText;
  //
  private final Settings settings;
  private final String rememberBaseKey;
  private final JComponent inputField;
  private final JButton filePicker;
  private static final long serialVersionUID = 1L;
}
