package com.ringlord.swing;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JPanel;


/**
 * <p>
 * A GridBagLayout-based form.
 * </p>
 * 
 * @author K Udo Schuermann
 **/
public class Form
  extends JPanel
{
  private static final long serialVersionUID = 1656902463225948277L;


  public Form()
  {
    this( new Insets( 1,
	              3,
	              1,
	              0 ) );
  }


  public Form( final Insets insets )
  {
    super( null );
    this.layout = new GridBagLayout();
    setOpaque( false );
    setLayout( layout );
    constraints.insets = insets;
  }


  public void addRow( final Component label,
	              final boolean northWestLabel,
	              final Component field )
  {
    if( northWestLabel )
      {
	final int row = autoRow;
	if( label != null )
	  {
	    add( row,
		 0,
		 label,
		 0.0,
		 0.0,
		 1,
		 1,
		 0,
		 0,
		 Sizing.NONE,
		 Anchor.NORTHWEST );
	  }
	add( row,
	     1,
	     field,
	     1.0,
	     0.0,
	     1,
	     1,
	     0,
	     0,
	     Sizing.WIDTH,
	     Anchor.WEST );
      }
    else
      {
	addRow( label,
	        field );
      }
  }


  public void addRow( final Component label,
	              final Component field )
  {
    addRow( label,
	    field,
	    null );
  }


  public void addRow( final Component field )
  {
    addRow( null,
	    field,
	    null );
  }


  public void addRow( final Component label,
	              final Component field,
	              final Component suffix )
  {
    addRow( label,
	    field,
	    1,
	    suffix );
  }


  /**
   * <p>
   * A convenience method that adds a 2 or 3 element row.
   * </p>
   * 
   * @param label
   *          An optional label component. The label component will be the
   *          left-most and will not resize when the parent is resized.
   * 
   * @param field
   *          An input field component that will be resized when the parent is
   *          resized. The 'span' parameter (see next) determines how many
   *          columns the field should span.
   * 
   * @param span
   *          The number of columns thatthe 'field' should span. This must be 1
   *          or greater. Most commonly this would be 1 or 2.
   * 
   * @param suffix
   *          An optional suffix component, perhaps additional text, or a button
   *          to open a color or file chooser dialog. The suffix component will
   *          not resize when the parent is resized.
   **/
  public void addRow( final Component label,
	              final Component field,
	              final int span,
	              final Component suffix )
  {
    final int row = autoRow;
    if( label != null )
      {
	add( row,
	     0,
	     label,
	     0.0,
	     0.0,
	     1,
	     1,
	     0,
	     0,
	     Sizing.NONE,
	     Anchor.WEST );
      }
    add( row,
	 1,
	 field,
	 1.0,
	 0.0,
	 1,
	 span,
	 0,
	 0,
	 Sizing.WIDTH,
	 Anchor.WEST );
    if( suffix != null )
      {
	add( row,
	     2,
	     suffix,
	     0.0,
	     0.0,
	     1,
	     1,
	     0,
	     0,
	     Sizing.NONE,
	     Anchor.WEST );
      }
  }


  /**
   * <p>
   * Place the given {@link Component} into the form's cell at (row,col). The
   * component will be aligned to the WEST (upper left) and will stretch in
   * width to fill available space.
   * </p>
   * 
   * <p>
   * The component's horizontal weight will be 1.0, but its vertical weight will
   * be zero. This allows the layout to stretch horizontally, but not vertically
   * (which would generally cause empty vertical space between rows).
   * </p>
   * 
   * <p>
   * Note that components need not be added in any specific order.
   * </p>
   * 
   * @param row
   *          The row, starting with 0.
   * @param col
   *          The column, starting with 0.
   * @param item
   *          The component (must not be null).
   **/
  public void add( final int row,
	           final int col,
	           final Component item )
  {
    add( row,
	 col,
	 item,
	 1.0,
	 0.0 );
  }


  /**
   * <p>
   * Place the given {@link Component} into the form's cell at (row,col), and
   * assigning it a specific horizontal and vertical weight.
   * </p>
   * 
   * <p>
   * Note that components need not be added in any specific order.
   * </p>
   * 
   * @param row
   *          The row, starting with 0.
   * @param col
   *          The column, starting with 0.
   * @param item
   *          The component (must not be null).
   * @param weightx
   *          The component's horizontal weight (0.0&ndash;1.0)
   * @param weighty
   *          The component's vertical weight (0.0&ndash;1.0)
   **/
  public void add( final int row,
	           final int col,
	           final Component item,
	           final double weightx,
	           final double weighty )
  {
    add( row,
	 col,
	 item,
	 weightx,
	 weighty,
	 1,
	 1 );
  }


  /**
   * <p>
   * Place the given {@link Component} into the form's cell at (row,col),
   * assigning it a specific horizontal and vertical weight, and potentially
   * spanning it across multiple columns and/or rows.
   * </p>
   * 
   * <p>
   * Note that components need not be added in any specific order.
   * </p>
   * 
   * @param row
   *          The row, starting with 0.
   * @param col
   *          The column, starting with 0.
   * @param item
   *          The component (must not be null).
   * @param weightx
   *          The component's horizontal weight (0.0&ndash;1.0)
   * @param weighty
   *          The component's vertical weight (0.0&ndash;1.0)
   * @param rowSpan
   *          The number of rows to cover (1+)
   * @param colSpan
   *          The number of columns to cover (1+)
   **/
  public void add( final int row,
	           final int col,
	           final Component item,
	           final double weightx,
	           final double weighty,
	           final int rowSpan,
	           final int colSpan )
  {
    add( row,
	 col,
	 item,
	 weightx,
	 weighty,
	 rowSpan,
	 colSpan,
	 0,
	 0,
	 Sizing.WIDTH,
	 Anchor.WEST );
  }


  /**
   * <p>
   * Place the given {@link Component} into the form's cell at (row,col),
   * assigning it a specific horizontal and vertical weight, potentially
   * spanning it across multiple columns and/or rows, stretching the cell by an
   * extra horizontal and/or vertical amount, and specifying sizing and anchor
   * positions.
   * </p>
   * 
   * <p>
   * Note that components need not be added in any specific order.
   * </p>
   * 
   * @param row
   *          The row, starting with 0.
   * @param col
   *          The column, starting with 0.
   * @param item
   *          The component (must not be null).
   * @param weightx
   *          The component's horizontal weight (0.0&ndash;1.0)
   * @param weighty
   *          The component's vertical weight (0.0&ndash;1.0)
   * @param rowSpan
   *          The number of rows to cover (1+)
   * @param colSpan
   *          The number of columns to cover (1+)
   * @param extrax
   *          The number of extra horizontal pixels to allocate.
   * @param extray
   *          The number of extra vertical pixels to allocate.
   * @param sizing
   *          How the component should be {@link #Sizing resized}
   * @param anchor
   *          How the component should be {@link #Anchor positioned}
   **/
  public void add( final int row,
	           final int col,
	           final Component item,
	           final double weightx,
	           final double weighty,
	           final int rowSpan,
	           final int colSpan,
	           final int extrax,
	           final int extray,
	           final Sizing sizing,
	           final Anchor anchor )
  {
    synchronized( content )
      {
	content.put( toID( row,
	                   col ),
	             new Item( item,
	                       row,
	                       col,
	                       rowSpan,
	                       colSpan,
	                       weightx,
	                       weighty,
	                       extrax,
	                       extray,
	                       sizing,
	                       anchor ) );
	autoRow = Math.max( row + 1,
	                    autoRow );
      }
    invalid = true;
  }


  @Override
  public Dimension getMinimumSize()
  {
    rebuildLayout();
    return super.getMinimumSize();
  }


  @Override
  public Dimension getPreferredSize()
  {
    return getMinimumSize();
  }


  public void clear()
  {
    synchronized( content )
      {
	if( !content.isEmpty() )
	  {
	    content.clear();
	    invalid = true;
	  }
      }
  }


  @Override
  public void remove( final Component item )
  {
    synchronized( content )
      {
	String toBeRemoved = null;
	for( final Map.Entry<String,Item> e : content.entrySet() )
	  {
	    if( e.getValue().item == item )
	      {
		toBeRemoved = e.getKey();
		break;
	      }
	  }
	if( toBeRemoved != null )
	  {
	    content.remove( toBeRemoved );
	    invalid = true;
	  }
      }
  }


  @Override
  public void paintComponent( final Graphics g )
  {
    rebuildLayout();
    super.paintComponent( g );
  }


  public enum Anchor
  {
    CENTER,
    NORTH,
    NORTHEAST,
    EAST,
    SOUTHEAST,
    SOUTH,
    SOUTHWEST,
    WEST,
    NORTHWEST;
  }


  public enum Sizing
  {
    NONE,
    WIDTH,
    HEIGHT,
    BOTH
  }


  private void rebuildLayout()
  {
    if( invalid )
      {
	synchronized( content )
	  {
	    removeAll();
	    final List<Item> items = new ArrayList<Item>();
	    for( final Item item : content.values() )
	      {
		items.add( item );
	      }
	    for( final Item item : items )
	      {
		if( !item.item.isVisible() )
		  {
		    continue;
		  }
		constraints.gridx = item.col;
		constraints.gridy = item.row;
		constraints.gridwidth = item.colSpan;
		constraints.gridheight = item.rowSpan;
		switch( item.sizing )
		  {
		  case NONE:
		    constraints.fill = GridBagConstraints.NONE;
		    break;

		  case WIDTH:
		    constraints.fill = GridBagConstraints.HORIZONTAL;
		    break;

		  case HEIGHT:
		    constraints.fill = GridBagConstraints.VERTICAL;
		    break;

		  case BOTH:
		    constraints.fill = GridBagConstraints.BOTH;
		    break;
		  }
		switch( item.anchor )
		  {
		  case CENTER:
		    constraints.anchor = GridBagConstraints.CENTER;
		    break;

		  case NORTH:
		    constraints.anchor = GridBagConstraints.PAGE_START;
		    break;

		  case NORTHEAST:
		    constraints.anchor = GridBagConstraints.FIRST_LINE_END;
		    break;

		  case EAST:
		    constraints.anchor = GridBagConstraints.LINE_END;
		    break;

		  case SOUTHEAST:
		    constraints.anchor = GridBagConstraints.LAST_LINE_END;
		    break;

		  case SOUTH:
		    constraints.anchor = GridBagConstraints.PAGE_END;
		    break;

		  case SOUTHWEST:
		    constraints.anchor = GridBagConstraints.LAST_LINE_START;
		    break;

		  case WEST:
		    constraints.anchor = GridBagConstraints.LINE_START;
		    break;

		  case NORTHWEST:
		    constraints.anchor = GridBagConstraints.FIRST_LINE_START;
		    break;
		  }
		constraints.weightx = item.weightx;
		constraints.weighty = item.weighty;
		constraints.ipadx = item.extrax;
		constraints.ipady = item.extray;

		layout.setConstraints( item.item,
		                       constraints );
		add( item.item );
	      }
	  }
	invalid = false;
      }
  }


  private static class Item
  {
    Item( final Component item,
	  final int row,
	  final int col,
	  final int rowSpan,
	  final int colSpan,
	  final double weightx,
	  final double weighty,
	  final int extrax,
	  final int extray,
	  final Sizing sizing,
	  final Anchor anchor )
    {
      super();
      this.item = item;
      this.row = row;
      this.col = col;
      this.rowSpan = rowSpan;
      this.colSpan = colSpan;
      this.weightx = weightx;
      this.weighty = weighty;
      this.extrax = extrax;
      this.extray = extray;
      this.sizing = sizing;
      this.anchor = anchor;
    }
    private final Component item;
    private final int row, col, rowSpan, colSpan, extrax, extray;
    private final double weightx, weighty;
    private final Sizing sizing;
    private final Anchor anchor;
  }


  private static String toID( final int row,
	                      final int col )
  {
    return (row + "/" + col);
  }

  private int autoRow = 0;
  private boolean invalid;
  private final GridBagLayout layout;
  private final GridBagConstraints constraints = new GridBagConstraints();
  private final Map<String,Item> content = new HashMap<String,Item>();
}
