package com.ringlord.swing;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.Component;
import java.awt.LayoutManager;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.List;
import java.util.ArrayList;
import java.util.EventObject;

import javax.swing.Icon;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JScrollPane;


/**
 * <p>
 * A "Wizard" container which manages the walk-through of a number of
 * {@link Page}S defined by the caller. Pages provide to the Guide the feedback
 * it needs to manage Back, Next/Finish (and Cancel) buttons. The Guide can
 * optionally skip initial Pages which report to the Guide that their data entry
 * is complete. Example: When configuring a database operation involving (among
 * other thigns) the URL, login name, and password) it may be useful to skip the
 * configuration page if previously completed choices continue to apply.
 * </p>
 * 
 * <p>
 * Please note that the Guide is not thread safe: You should not add/remove
 * {@link Page}S from multiple threads, nor add/remove Pages while the Guide is
 * visible (in operation).
 * </p>
 * 
 * <p>
 * Tip: Traditionally, "Wizards" feature an introductory/welcome page as well as
 * a final confirmation page. If you follow this practice, both of these should
 * always report as being ready to proceed, which allows the user to simply
 * press "Next"/"Finish".
 * </p>
 * 
 * <p>
 * The following <tt>TestGuide</tt> class demonstrates the use of the Guide:
 * </p>
 * 
 * <pre>
 * import java.awt.event.ActionListener;
 * import java.awt.event.ActionEvent;
 * import javax.swing.JDialog;
 * 
 * import com.ringlord.swing.Guide;
 * 
 * 
 * public class TestGuide
 * {
 *   public static void main( String[] args )
 *   {
 *     Guide guide = new Guide();
 *     guide.addPage( new TestPage( &quot;Welcome Page&quot;,
 * 	                         true ) );
 *     guide.addPage( new TestPage( &quot;Entry Part 1/2&quot; ) );
 *     guide.addPage( new TestPage( &quot;Entry Part 2/2&quot; ) );
 *     guide.addPage( new TestPage( &quot;Confirmation Page&quot;,
 * 	                         true ) );
 * 
 *     final JDialog d = new JDialog( (JFrame)null );
 * 
 *     // The ActionListener manipulates the dialog, and
 *     // for this to be possible the dialog (d) must be
 *     // a final variable.
 *     guide.addActionListener( new ActionListener()
 *     {
 *       public void actionPerformed( final ActionEvent e )
 *       {
 * 	if( guide.isFinished( e ) )
 * 	  {
 * 	    d.setVisible( false ); // hide
 * 	    d.dispose(); // allow GUI to exit
 * 	    // User has finished, begin processing
 * 	  }
 * 	else if( guide.isCanceled( e ) ) // must be the case
 * 	  {
 * 	    d.setVisible( false );
 * 	    d.dispose();
 * 	    // User has canceled, do not process
 * 	  }
 *       }
 *     } );
 * 
 *     d.add( guide );
 *     d.pack();
 *     d.setVisible( true );
 *   }
 * }
 * </pre>
 * 
 * <p>
 * The following is a fully functioning <tt>TestPage</tt> class, as used by the
 * sample code above:
 * </p>
 * 
 * <pre>
 * import java.awt.event.ItemListener;
 * import java.awt.event.ItemEvent;
 * import javax.swing.JCheckBox;
 * 
 * import com.ringlord.swing.Guide;
 * 
 * 
 * class TestPage
 *   extends Guide.Page
 *   implements
 *     ItemListener
 * {
 *   public TestPage( final String name )
 *   {
 *     this( name,
 * 	  false );
 *   }
 * 
 * 
 *   public TestPage( final String name,
 * 	           final boolean isReady )
 *   {
 *     this.name = name;
 *     // Here we build the GUI which consists of a single
 *     // JCheckBox. When this is checked, the page is
 *     // considered ready to proceed (the Guide's &quot;Next&quot;
 *     // or &quot;Finish&quot; button will become active).
 *     this.x = new JCheckBox( &quot;All the work is done for &quot; + name,
 * 	                    isReady );
 *     this.x.addItemListener( this );
 *     add( this.x );
 *   }
 * 
 * 
 *   public void itemStateChanged( final ItemEvent e )
 *   {
 *     // The user has clicked the JCheckBox, either checking
 *     // or unchecking it. In this example this is the sole
 *     // determinant whether data collection is done, and we
 *     // notify the Guide (using a setReady(true/false) call)
 *     // of our ready state:
 *     setReady( x.isSelected() );
 *   }
 * 
 * 
 *   public boolean isOK()
 *   {
 *     if( isFirst )
 *       {
 * 	// We could perform first-time initialization here,
 * 	// but if this Page's data depends on other page's
 * 	// data then we might need to update things at
 * 	// other times, too.
 * 	isFirst = false;
 *       }
 *     return x.isSelected();
 *   }
 *   //
 *   private final javax.swing.JCheckBox x;
 *   private boolean isFirst = true;
 *   private String name;
 * }
 * </pre>
 * 
 * @author K Udo Schuermann
 **/
public class Guide
  extends JPanel
{
  /**
   * 
   */
  private static final long serialVersionUID = 1L;


  /**
   * Construct a Guide which has no {@link Page}S initially.
   **/
  public Guide()
  {
    this( false );
  }


  public Guide( final boolean skipReadyPages )
  {
    this( skipReadyPages,
	  null );
  }


  /**
   * Construct a Guide which has no {@link Page}S initially.
   * 
   * @param skipReadyPages
   **/
  public Guide( final boolean skipReadyPages,
	        final Icon[] iconsBackNextFinishCancel )
  {
    super( new BorderLayout() );
    this.skipReadyPages = skipReadyPages;
    this.buttonBar = new JPanel( new FlowLayout( FlowLayout.RIGHT ) );
    this.pages = new ArrayList<Page>();
    this.curPageIndex = -1;
    this.controller = new PageController();

    final Icon back, next, finish, cancel;
    if( iconsBackNextFinishCancel != null )
      {
	back = (iconsBackNextFinishCancel.length > 0
	    ? iconsBackNextFinishCancel[0]
	    : null);
	next = (iconsBackNextFinishCancel.length > 1
	    ? iconsBackNextFinishCancel[1]
	    : null);
	finish = (iconsBackNextFinishCancel.length > 2
	    ? iconsBackNextFinishCancel[2]
	    : null);
	cancel = (iconsBackNextFinishCancel.length > 3
	    ? iconsBackNextFinishCancel[3]
	    : null);
      }
    else
      {
	back = Images.getIcon( "/img/icon_back.png" );
	next = Images.getIcon( "/img/icon_next.png" );
	finish = Images.getIcon( "/img/icon_accept.png" );
	cancel = Images.getIcon( "/img/icon_cancel.png" );
      }

    final JPanel buttonGrid = new JPanel( new GridLayout( 1,
	                                                  0 ) );
    buttonGrid.add( backButton = new JButton( "Back",
	                                      back ) );
    buttonGrid.add( nextButton = new JButton( "Next",
	                                      next ) );
    buttonGrid.add( finishButton = new JButton( "Finish",
	                                        finish ) );
    buttonGrid.add( cancelButton = new JButton( "Cancel",
	                                        cancel ) );
    buttonBar.add( buttonGrid );

    this.pagePanel = new JPanel( cards = new CardLayout() );

    add( new JScrollPane( pagePanel ),
	 BorderLayout.CENTER );
    add( buttonBar,
	 BorderLayout.SOUTH );
    isInitializing = false; // can only add Page objects from here on out

    backButton.setEnabled( false );
    nextButton.setEnabled( false );
    finishButton.setVisible( false );

    backButton.addActionListener( new ActionListener()
    {
      public void actionPerformed( final ActionEvent e )
      {
	cards.previous( pagePanel );
	--curPageIndex;
	updateButtons();
      }
    } );
    nextButton.addActionListener( new ActionListener()
    {
      public void actionPerformed( final ActionEvent e )
      {
	cards.next( pagePanel );
	++curPageIndex;
	updateButtons();
      }
    } );
    final Guide guide = this; // ActionListenerS need a final reference to this
	                      // (Guide)
    finishButton.addActionListener( new ActionListener()
    {
      public void actionPerformed( final ActionEvent e )
      {
	final ActionEvent event = new ActionEvent( guide,
	                                           1,
	                                           "FINISHED" );
	for( ActionListener l : actionListeners )
	  {
	    l.actionPerformed( event );
	  }
      }
    } );
    cancelButton.addActionListener( new ActionListener()
    {
      public void actionPerformed( final ActionEvent e )
      {
	final ActionEvent event = new ActionEvent( guide,
	                                           0,
	                                           "CANCELED" );
	for( ActionListener l : actionListeners )
	  {
	    l.actionPerformed( event );
	  }
      }
    } );
  }


  // ----------------------------------------------------------------------

  /**
   * <p>
   * Add an {@link ActionListener} to the Guide. The Guide reports to the
   * listeners either "FINISHED" or "CANCELED"
   * {@link ActionListener#getActionCommand()}, but it is recommended to call
   * {@link #isFinished(ActionEvent)} and {@link #isCanceled(ActionEvent)}
   * instead of testing for the action text direction.
   * </p>
   **/
  public void addActionListener( final ActionListener l )
  {
    actionListeners.add( l );
  }


  public void removeActionListener( final ActionListener l )
  {
    actionListeners.remove( l );
  }


  // ----------------------------------------------------------------------

  /**
   * Test whether the ActionEvent provided by the Guide to an ActionListener
   * indicates that the user has pressed the "Finish" button.
   **/
  public boolean isFinished( final ActionEvent e )
  {
    return "FINISHED".equals( e.getActionCommand() );
  }


  /**
   * Test whether the ActionEvent provided by the Guide to an ActionListener
   * indicates that the user has pressed the "Cancel" button.
   **/
  public boolean isCanceled( final ActionEvent e )
  {
    return "CANCELED".equals( e.getActionCommand() );
  }


  // ----------------------------------------------------------------------

  @Override
  public void add( final Component c,
	           final Object x )
  {
    if( isInitializing )
      {
	super.add( c,
	           x );
      }
    else
      {
	addPage( (Page)c );
      }
  }


  @Override
  public void add( final Component c,
	           final Object x,
	           final int n )
  {
    if( isInitializing )
      {
	super.add( c,
	           x,
	           n );
      }
    else
      {
	addPage( (Page)c,
	         n );
      }
  }


  @Override
  public Component add( final Component c )
  {
    if( isInitializing )
      {
	return super.add( c );
      }
    else
      {
	return addPage( (Page)c );
      }
  }


  @Override
  public Component add( final String s,
	                final Component x )
  {
    if( isInitializing )
      {
	return super.add( s,
	                  x );
      }
    else
      {
	return addPage( (Page)x );
      }
  }


  @Override
  public Component add( final Component x,
	                final int n )
  {
    if( isInitializing )
      {
	return super.add( x,
	                  n );
      }
    else
      {
	return addPage( (Page)x,
	                n );
      }
  }


  @Override
  public void remove( final int n )
  {
    removePage( pages.get( n ) );
  }


  @Override
  public void remove( final Component c )
  {
    removePage( (Page)c );
  }


  /**
   * Adds a {@link Page} to the Guide.
   **/
  public Page addPage( final Page page )
  {
    return addPage( page,
	            pages.size() );
  }


  /**
   * <p>
   * Adds a {@link Page} to the Guide at a specific position.
   * </p>
   * 
   * <p>
   * There are issues with inserting pages before others if the Guide has been
   * configured to auto skip ready pages.
   * </p>
   **/
  public Page addPage( final Page page,
	               final int n )
  {
    pages.add( n,
	       page );
    page.addPageStateListener( controller );
    pagePanel.add( page,
	           n );
    if( curPageIndex < 0 )
      {
	pagePanel.setVisible( true );
	curPageIndex = 0;
	cards.first( pagePanel );
      }
    else if( n == 0 )
      {
	cards.first( pagePanel );
      }

    if( skipReadyPages && ((curPageIndex + 2) == pages.size()) )
      {
	Page p = pages.get( curPageIndex );
	if( p.isOK() )
	  {
	    ++curPageIndex;
	    cards.next( pagePanel );
	  }
      }

    updateButtons();
    return page;
  }


  public void removePage( final Page page )
  {
    final int pageIndex = pages.indexOf( page );
    if( pageIndex >= 0 )
      {
	pagePanel.remove( page );
	if( curPageIndex >= pageIndex )
	  {
	    curPageIndex = pageIndex - 1;
	    updateButtons();
	  }
      }
  }


  @Override
  public void removeAll()
  {
    removeAllPages();
  }


  /**
   * Removes all {@link Page}S from the Guide.
   **/
  public void removeAllPages()
  {
    pages.clear();
    if( curPageIndex >= 0 )
      {
	curPageIndex = -1;
	updateButtons();
      }
  }


  /**
   * Updates the state of all buttons.
   **/
  private void updateButtons()
  {
    if( curPageIndex < 0 )
      {
	backButton.setVisible( false );
	nextButton.setVisible( false );
	finishButton.setVisible( false );
      }
    else
      {
	final Page page = pages.get( curPageIndex );

	backButton.setVisible( true );
	backButton.setEnabled( curPageIndex > 0 );

	if( (curPageIndex + 1) < pages.size() )
	  {
	    nextButton.setVisible( true );
	    nextButton.setEnabled( page.isOK() );

	    finishButton.setVisible( false );
	  }
	else
	  {
	    nextButton.setVisible( false );

	    finishButton.setVisible( true );
	    finishButton.setEnabled( page.isOK() );
	  }
      }
  }


  // ======================================================================

  /**
   * <p>
   * A {@link JPanel}
   **/
  public abstract static class Page
    extends JPanel
  {
    public Page()
    {
      super( new BorderLayout() );
    }


    public Page( final LayoutManager layout )
    {
      super( layout );
    }


    /**
     * <p>
     * Indicates whether the page is ready to proceed or not.
     * </p>
     * 
     * <p>
     * This method is called whenever the page is made active (moving either
     * forward or backwards through the set of pages) and the state of the
     * "Next"/"Finish" buttons must be set.
     * </p>
     * 
     * <p>
     * The first time it is called is a good time to initialize default values,
     * but you may need to alter the page state based on changes that were made
     * to other pages before this page is being displayed.
     * </p>
     * 
     * @return 'true' if the page is complete and the user should have the
     *         option of proceeding to the next page (or click finish).
     **/
    public abstract boolean isOK();


    /**
     * <p>
     * Notify the Guide of the Page's new state, either ready to proceed or not,
     * allowing the Guide to enable/disable the state of the "Next"/"Finish"
     * buttons.
     * </p>
     * 
     * <p>
     * This method should be called in direct response to user actions that
     * alter the page's readiness state.
     * </p>
     * 
     * @param isReady
     *          Indicate whether the page is complete and the user may proceed
     *          to the next page (or click finish if this is the last page).
     **/
    protected final void setReady( final boolean isReady )
    {
      final PageStateEvent e = new PageStateEvent( this,
	                                           (isReady
	                                               ? PageStateEvent.Type.READY
	                                               : PageStateEvent.Type.NEEDY) );
      notifyPageStateListeners( e );
    }


    /**
     * Invoked by the {@link #setReady} method. Not public.
     **/
    private final void notifyPageStateListeners( final PageStateEvent e )
    {
      for( PageStateListener l : psListeners )
	{
	  l.pageStateChanged( e );
	}
    }


    /**
     * Invoked by the Guide when a page is added, in order to listen to the
     * Page's state as it changes due to user action.
     **/
    private final void addPageStateListener( final PageStateListener l )
    {
      psListeners.add( l );
    }


    /**
     * Invoked by the Guide when a page is removed.
     **/
    private final void removePageStateListener( final PageStateListener l )
    {
      psListeners.remove( l );
    }
    private final List<PageStateListener> psListeners = new ArrayList<PageStateListener>();
    private static final long serialVersionUID = 1L;
  }


  // ======================================================================

  /**
   * Controls the state of the Back and Next/Finish buttons and the movement
   * through the individual pages. This class is internal to the Guide.
   * 
   * @author K Udo Schuermann
   **/
  private class PageController
    implements
      PageStateListener
  {
    public void pageStateChanged( final PageStateEvent e )
    {
      final Page page = e.getPage();
      final PageStateEvent.Type type = e.getType();

      curPageIndex = pages.indexOf( page );
      switch( type )
	{
	case READY:
	  if( (curPageIndex + 1) < pages.size() )
	    {
	      nextButton.setEnabled( true );
	    }
	  else
	    {
	      finishButton.setEnabled( true );
	    }
	  break;

	case NEEDY:
	  if( (curPageIndex + 1) < pages.size() )
	    {
	      nextButton.setEnabled( false );
	    }
	  else
	    {
	      finishButton.setEnabled( false );
	    }
	  break;

	default:
	  System.err.println( "Unexpected page state event signal: " + type );
	}
    }
  }


  // ======================================================================

  /**
   * An event fired by a Page object to indicate whether it is ready to proceed.
   * This class is internal to the Guide.
   * 
   * @author K Udo Schuermann
   **/
  private static class PageStateEvent
    extends EventObject
  {
    /**
     * 
     */
    private static final long serialVersionUID = 1L;


    PageStateEvent( final Page source,
	            final Type type )
    {
      super( source );
      this.type = type;
    }


    public Page getPage()
    {
      return (Page)getSource();
    }


    public Type getType()
    {
      return type;
    }


    public enum Type
    {
      READY,
      NEEDY
    }
    private Type type;
  }


  // ======================================================================

  /**
   * Defines the interface for a listener interested in the state changes of a
   * Page, usually in order to control the availability of the Next/Finish
   * buttons. This class is internal to the Guide.
   * 
   * @author K Udo Schuermann
   **/
  private interface PageStateListener
  {
    public void pageStateChanged( final PageStateEvent e );
  }

  // ======================================================================

  private boolean isInitializing = true;
  private int curPageIndex = -1;
  private final CardLayout cards;
  private final JPanel pagePanel;
  private final JPanel buttonBar;
  private final JButton backButton, nextButton, finishButton, cancelButton;
  private final PageStateListener controller;
  private final boolean skipReadyPages;
  private final List<Page> pages;
  private final List<ActionListener> actionListeners = new ArrayList<ActionListener>();
}
