package com.ringlord.swing;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JMenu;
import javax.swing.JMenuItem;


/**
 * <p>
 * A JMenu that provides for a limited set of "Recently used" entries. Selecting
 * an entry from the list fires the ActionEvent for the JMenu after re-arranging
 * the elements to that the selected element now comes first in the list.
 * </p>
 * 
 * <p>
 * When the ActionEvent arrives, call {@link #matchItem(JMenuItem)} to determine
 * the original element type associated with that menu. In other words, if you
 * declare a JRecentMenu&lt;File&gt; and the ActionListener you have attached to
 * this JRecentMenu (prior to adding Recent elements) receives an ActionEvent,
 * then the {@link ActionEvent#getSource()} will be a JMenuItem which you should
 * pass to {@link #matchItem(JMenuItem)} to obtain the Recent (File) element.
 * </p>
 * 
 * <p>
 * Note: In order for {@link ActionEvent}S to reach you, add an ActionListener
 * to the JRecentMenu <em>before calling {@link #setRecent(List)} or
 * {@link #addRecent(Object)}</em>, because the Recent elements will inherit all
 * (or none) of the ActionListenerS from the JRecentMenu.
 * </p>
 * 
 * @author K Udo Schuermann
 */
public class JRecentMenu<T>
  extends JMenu
{
  /**
   * Creates a simple JRecentMenu named "Recent", limited to 10 entries, and
   * without offering the user the ability to clear the list.
   */
  public JRecentMenu()
  {
    this( "Recent",
	  10,
	  null );
  }


  /**
   * Create a JRecentMenu
   * 
   * @param title
   *          The title for the JMenu
   * @param maxEntryCount
   *          The maximum number of entries. When a new entry is
   *          {@link #addRecent added}, or more than this many entries are
   *          {@link #setRecent(List)}, the overflow is discarded.
   * @param clearName
   *          An optional text for a menu entry added at the end of the list
   *          that allows the user the ability to clear the list. If null, no
   *          such entry is produced.
   */
  public JRecentMenu( final String title,
	              final int maxEntryCount,
	              final String clearName )
  {
    super( title );
    this.maxEntryCount = maxEntryCount;
    if( clearName == null )
      {
	this.clearMenu = null;
      }
    else
      {
	this.clearMenu = new JMenuItem( clearName );
	this.clearMenu.addActionListener( new ActionListener()
	{
	  @Override
	  public void actionPerformed( final ActionEvent e )
	  {
	    removeAllRecent();
	  }
	} );
      }
  }


  /**
   * Call this method with the {@link ActionEvent#getSource()} (no need to cast
   * it to anything) to determine the Recent element (of type T) that was
   * selected. Note that you need (and should) not rearrange any elements, the
   * selected element will be (or already has been) moved to first place in the
   * menu.
   * 
   * @param menuEntry
   *          A JMenuItem from this menu; any object not a JMenuItem on this
   *          JMenu will simply produce no match.
   * @return The Recent element associated with the given menuEntry or null if
   *         no match was found.
   */
  public T matchItem( final Object menuEntry )
  {
    for( final Entry entry : items )
      {
	if( entry.getJMenuItem() == menuEntry )
	  {
	    return entry.getItem();
	  }
      }
    return null;
  }


  /**
   * Obtain a list of all Recent elements in the order in which they are
   * presently displayed. The same list can be passed to
   * {@link #setRecent(List)} to restore the menu.
   * 
   * @return A list of zero or more Recent elements.
   * @see #setRecent(List)
   * @see #addRecent(T);
   */
  public synchronized List<T> getAllRecent()
  {
    final List<T> result = new ArrayList<>();
    for( final Entry entry : items )
      {
	result.add( entry.getItem() );
      }
    return result;
  }


  /**
   * Sets the Recent items in the menu.
   * 
   * @param items
   */
  public synchronized void setRecent( final List<T> items )
  {
    removeAllRecent();
    if( items != null )
      {
	final ActionListener[] actionListeners = getActionListeners();
	for( final T item : items )
	  {
	    if( item != null )
	      {
		this.items.add( new Entry( item,
		                           actionListeners ) );
		if( this.items.size() >= maxEntryCount )
		  {
		    break;
		  }
	      }
	  }
	rebuildMenu();
      }
  }


  /**
   * Adds one Recent element to the top of the list. If the element already
   * exists in the list, it is moved to the top (if not already at first
   * position).
   * 
   * @param item
   *          The Recent element to be added.
   */
  public synchronized void addRecent( final T item )
  {
    if( item != null )
      {
	final int entryCount = items.size();
	for( int i = 0; i < entryCount; i++ )
	  {
	    final Entry e = items.get( i );
	    if( e.getItem().equals( item ) )
	      {
		if( i > 0 ) // not already at first place?
		  {
		    items.remove( i );
		    items.add( 0,
			       e );
		    rebuildMenu();
		  }
		return;
	      }
	  }

	final Entry e = new Entry( item,
	                           getActionListeners() );
	items.add( 0,
	           e );
	while( this.items.size() >= maxEntryCount )
	  {
	    final Entry overflow = this.items.remove( this.items.size() - 1 );
	    clean( overflow );
	  }
	rebuildMenu();
      }
  }


  /**
   * Remove all Recent elements.
   */
  public synchronized void removeAllRecent()
  {
    if( items.size() > 0 )
      {
	for( final Entry e : items )
	  {
	    clean( e );
	  }
	removeAll();
	items.clear();
      }
    rebuildMenu();
  }


  /**
   * Rebuilds the menu
   */
  private void rebuildMenu()
  {
    removeAll();
    for( final Entry entry : items )
      {
	add( entry.getJMenuItem() );
      }
    if( (items.size() > 0) && (clearMenu != null) )
      {
	addSeparator();
	add( clearMenu );
      }
  }


  /**
   * Cleans an {@link Entry} (removes ActionListeners)
   * 
   * @param e
   */
  private void clean( final Entry e )
  {
    final JMenuItem menuItem = e.getJMenuItem();
    final ActionListener[] actionListeners = menuItem.getActionListeners();
    for( final ActionListener l : actionListeners )
      {
	e.getJMenuItem().removeActionListener( l );
      }
  }


  private class Entry
    implements
      ActionListener
  {
    Entry( final T item,
	   final ActionListener[] actionListeners )
    {
      super();
      this.item = item;
      this.menuItem = new JMenuItem( item.toString() );
      menuItem.addActionListener( this );
      for( final ActionListener l : actionListeners )
	{
	  menuItem.addActionListener( l );
	  System.out.println( "Added ActionListener to " + item );
	}
    }


    T getItem()
    {
      return item;
    }


    JMenuItem getJMenuItem()
    {
      return menuItem;
    }


    @Override
    public void actionPerformed( final ActionEvent e )
    {
      final T item = matchItem( e.getSource() );
      if( item != null )
	{
	  addRecent( item );
	}
    }

    private final T item;
    private final JMenuItem menuItem;
  }

  private final int maxEntryCount;
  private final JMenuItem clearMenu;
  private final List<Entry> items = new ArrayList<>();
  private static final long serialVersionUID = -7726650810680095587L;
}
