package com.ringlord.swing;

import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.font.FontRenderContext;
import java.awt.font.LineMetrics;
import java.awt.font.TextLayout;
import java.awt.geom.Rectangle2D;
import java.net.URI;
import java.util.HashSet;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.border.Border;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;


/**
 * <p>
 * A Link an external (by {@link URI})-identified resource, and intended to be
 * {@linkplain TextEditor#insertLink(Link) inserted in a TextEditor} (as a
 * custom-rendered {@link JButton} you can also add it to any JTextPane using
 * the insertComponent method).
 * </p>
 * 
 * <p>
 * Because JTextPanel fails to align baselines of embedded components, aligning
 * the component centerline with the surrounding text's baseline, instead, the
 * Link currently implements various hacks and kludges to make things look
 * right, but this is known to work correctly only for MonoSpace 10\u201313pt fonts,
 * but even so it may not even work across different platforms, or with
 * different versions of the Swing library.
 * <em>For this reason, this class should be considered a work-in-progress.</em>
 * </p>
 * 
 * @author K Udo Schuermann
 */
public class Link
  extends JButton
  implements
    Icon,
    MouseListener
{
  /**
   * Construct a URI Link.
   * 
   * @param uri
   *          The URI that an {@link ActionListener} can retrieve from the Link
   *          using {@link #getURI()}.
   * @param text
   *          The text to be rendered for this Link. It is accessible via
   *          {@link #getText()}.
   */
  public Link( final URI uri,
	       final String text )
  {
    super( text );

    this.changeListeners = new HashSet<>();
    setURI( uri );
    setText( text );

    addMouseListener( this );
    setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
  }


  /**
   * Update the URI associated with this Link.
   * 
   * @param uri
   *          The new URI
   */
  public void setURI( final URI uri )
  {
    this.uri = uri;
    fireChangeEvent( new ChangeEvent( this ) );
  }


  /**
   * Obtains the {@link URI} associated with this Link.
   * 
   * @return The URI
   */
  public URI getURI()
  {
    return uri;
  }


  @Override
  public void setFont( final Font font )
  {
    super.setFont( font );
    fireChangeEvent( new ChangeEvent( this ) );

    border = null;
  }


  @Override
  public void setText( final String text )
  {
    super.setText( text );
    fireChangeEvent( new ChangeEvent( this ) );
  }


  public void addChangeListener( final ChangeListener l )
  {
    if( changeListeners == null )
      {
	changeListeners = new HashSet<>();
      }
    changeListeners.add( l );
  }


  public void removeChangeListener( final ChangeListener l )
  {
    if( changeListeners == null )
      {
	changeListeners = new HashSet<>();
      }
    changeListeners.remove( l );
  }


  private void fireChangeEvent( final ChangeEvent e )
  {
    if( changeListeners != null )
      {
	for( ChangeListener l : changeListeners )
	  {
	    l.stateChanged( e );
	  }
      }
  }


  @Override
  public int getIconHeight()
  {
    return dim.height;
  }


  @Override
  public int getIconWidth()
  {
    return dim.width;
  }


  @Override
  public void paintIcon( final Component c,
	                 final Graphics g,
	                 final int x,
	                 final int y )
  {
    render( g,
	    x,
	    y );
  }


  @Override
  public void paintComponent( final Graphics g )
  {
    // The following helped figure out the bounds of the component.
    // NOTE: We need to support highlighting as part of text selection,
    // so this is how we fill the background, though we probably want
    // to avoid filling the full height.
    //
    // final Graphics2D g2 = (Graphics2D)g;
    // g2.setColor( Color.yellow );
    // g2.fill( g2.getClip() );

    // JTextPane aligns components against the JTextPane's text using the
    // component's CENTER, TOP, or BOTTOM line rather than the component
    // text's own baseline.
    paintIcon( this,
	       g,
	       0,
	       -2 );
  }


  /**
   * Called by paintIcon, which is called by paintComponent, so that the Link
   * can operate as an Icon as well as a Component.
   * 
   * @param g
   * @param x
   * @param y
   */
  private void render( final Graphics g,
	               final int x,
	               final int y )
  {
    final Graphics2D g2 = (Graphics2D)g;
    g2.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING,
	                 RenderingHints.VALUE_TEXT_ANTIALIAS_ON );

    final FontRenderContext frc = g2.getFontRenderContext();

    if( border == null )
      {
	final LineMetrics m = getFont().getLineMetrics( "hg",
	                                                frc );
	final float vOfs = m.getDescent();
	border = BorderFactory.createEmptyBorder( (int)(vOfs * 2),
	                                          0,
	                                          0,
	                                          0 );
	setBorder( border );
      }

    final String text = getText();
    TextLayout layout = new TextLayout( text,
	                                getFont(),
	                                frc );
    final Rectangle2D rect = layout.getBounds();
    final int w = (int)Math.ceil( rect.getWidth() ) + 1; // +1 word spacing
    final int h = (int)Math.ceil( rect.getHeight() );
    if( (w != dim.width) || (h != dim.height) )
      {
	dim = new Dimension( w,
	                     h );
	final java.awt.font.LineMetrics metrics = getFont().getLineMetrics( text,
	                                                                    frc );
	uOffset = metrics.getUnderlineOffset();
	ascent = metrics.getAscent();
	revalidate();
      }
    else
      {
	if( isMouseHovering && isMouseDown )
	  {
	    g2.setColor( Color.red );
	  }
	else
	  {
	    g2.setColor( Color.blue );
	  }
	layout.draw( g2,
	             x + 0,
	             y + ascent + 1 );
	if( isMouseHovering )
	  {
	    final int vOfs = y + (int)(ascent + uOffset) + 1;
	    g2.drawLine( x,
		         vOfs,
		         x + w,
		         vOfs );
	  }
      }
  }


  @Override
  public void mouseEntered( final MouseEvent e )
  {
    isMouseHovering = true;
    repaint();
  }


  public void mouseExited( final MouseEvent e )
  {
    isMouseHovering = false;
    repaint();
  }


  @Override
  public void mouseClicked( final MouseEvent e )
  {
  }


  @Override
  public void mousePressed( final MouseEvent e )
  {
    isMouseDown = true;
    repaint();
  }


  @Override
  public void mouseReleased( final MouseEvent e )
  {
    isMouseDown = false;
    revalidate();
    repaint();
  }

  private Dimension dim = new Dimension( 1,
	                                 1 );
  private float uOffset;
  private float ascent;
  private Border border;
  private boolean isMouseHovering;
  private boolean isMouseDown;
  //
  private URI uri;
  //
  private Set<ChangeListener> changeListeners;
  //
  private static final long serialVersionUID = 4184868747886984491L;
}