package com.ringlord.swing;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;

import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.border.Border;
import javax.swing.plaf.basic.BasicSplitPaneDivider;
import javax.swing.plaf.basic.BasicSplitPaneUI;


/**
 * <p>
 * A Component reminiscent of, and based on the {@link javax.swing.JSplitPane},
 * except that it supports more than two child components, and can therefore
 * serve to create more complex layout easily.
 * </p>
 * 
 * <p>
 * All components added to a MultiSplitPane combine the {@link Component} with a
 * {@link MultiSplitPane.Element#resizeWeight()} using an
 * {@link MultiSplitPane.Element}.
 * </p>
 * 
 * @author K Udo Schuermann
 */
public class MultiSplitPane
  extends JPanel
{
  public MultiSplitPane( final Element... elements )
  {
    this( SplitOrientation.HORIZONTAL,
	  true,
	  elements );
  }


  public MultiSplitPane( final SplitOrientation direction,
	                 final boolean continuousLayout,
	                 final Element... elements )
  {
    super( new BorderLayout() );
    this.orientation = direction;
    this.continuousLayout = continuousLayout;
    setElements( elements );
  }


  public void setElements( final Element... elements )
  {
    if( elements == null )
      {
	this.elements = new Element[0];
      }
    else
      {
	this.elements = new Element[elements.length];
	for( int i = 0; i < elements.length; i++ )
	  {
	    this.elements[i] = elements[i].clone();
	  }
      }
    invalid = true;
  }


  public void setDividerSize( final int dividerSize )
  {
    this.dividerSize = dividerSize;
  }


  public void setAreaWeights( final double[] areaWeights )
  {
    double total = 0.0d;
    for( int i = 0; i < areaWeights.length; i++ )
      {
	total += areaWeights[i];
      }

    if( total > 0.0d )
      {
	for( int i = 0; i < elements.length; i++ )
	  {
	    if( i < this.elements.length )
	      {
		this.elements[i].setRelativeSize( areaWeights[i] / total );
	      }
	  }

	invalid = true;
	repaint();
      }
  }


  public double[] getAreaWeights()
  {
    final double[] weights = new double[this.elements.length];
    int totalSize = 0;
    for( int i = 0; i < weights.length; i++ )
      {
	final Rectangle bounds = this.elements[i].component.getBounds();
	final int dimension = (orientation == SplitOrientation.HORIZONTAL
	    ? bounds.width
	    : bounds.height);

	weights[i] = dimension;
	totalSize += dimension;
      }
    for( int i = 0; i < weights.length; i++ )
      {
	weights[i] = weights[i] / totalSize;
      }
    return weights;
  }


  public Element[] elements()
  {
    final Element[] elements = new Element[this.elements.length];
    for( int i = 0; i < elements.length; i++ )
      {
	elements[i] = this.elements[i].clone();
      }
    return elements;
  }


  @Override
  public void paintComponent( final Graphics g )
  {
    rebuildLayout();
    super.paintComponent( g );
  }


  public void setDividerColor( final Color dividerColor )
  {
    this.dividerColor = dividerColor;
  }


  private synchronized void rebuildLayout()
  {
    if( invalid )
      {
	super.removeAll();

	final Rectangle bounds = getBounds();
	final int totalSize = (orientation == SplitOrientation.HORIZONTAL
	    ? bounds.width
	    : bounds.height);

	// If any elements lacked a specific size then assign them a size that
	// spaces them evenly
	for( int i = 0; i < elements.length; i++ )
	  {
	    if( !elements[i].isSized() )
	      {
		elements[i].setRelativeSize( 1.0d / elements.length );
	      }
	  }

	JSplitPane p = null;
	for( int i = 0; i < elements.length; i++ )
	  {
	    if( i == 0 )
	      {
		// Create the initial component and add it to the layout
		p = new JSplitPane( orientation.getJSplitPaneOrientation(),
		                    continuousLayout );
		p.setBackground( Color.BLACK );
		p.setUI( new BasicSplitPaneUI()
		{
		  @Override
		  public BasicSplitPaneDivider createDefaultDivider()
		  {
		    return new BasicSplitPaneDivider( this )
		    {
		      @Override
		      public void setBorder( final Border b )
		      {
			// do nothing
		      }


		      @Override
		      public void paint( final Graphics g )
		      {
			g.setColor( dividerColor );
			final Dimension d = getSize();
			g.fillRect( 0,
			            0,
			            d.width,
			            d.height );
		      }
		      private static final long serialVersionUID = 7826209942007617869L;
		    };
		  }
		} );
		p.setBorder( null );
		p.setDividerSize( dividerSize );
		add( BorderLayout.CENTER,
		     p );
		p.setLeftComponent( elements[i].component() );
		p.setResizeWeight( elements[i].resizeWeight() );
		final int thisSize = (int)(elements[i].getRelativeSize() * totalSize);
		p.setDividerLocation( thisSize );
	      }
	    else
	      {
		if( (i + 1) < elements.length )
		  {
		    // We have at least 2 more elements
		    final JSplitPane q = new JSplitPane( orientation.getJSplitPaneOrientation(),
			                                 continuousLayout );
		    q.setBackground( Color.BLACK );
		    q.setUI( new BasicSplitPaneUI()
		    {
		      @Override
		      public BasicSplitPaneDivider createDefaultDivider()
		      {
			return new BasicSplitPaneDivider( this )
			{
			  @Override
			  public void setBorder( final Border b )
			  {
			    // do nothing
			  }


			  @Override
			  public void paint( final Graphics g )
			  {
			    g.setColor( dividerColor );
			    final Dimension d = getSize();
			    g.fillRect( 0,
				        0,
				        d.width,
				        d.height );
			  }
			  private static final long serialVersionUID = -8351730427122268991L;
			};
		      }
		    } );
		    q.setDividerSize( dividerSize );
		    q.setBorder( null );
		    q.setLeftComponent( elements[i].component() );
		    q.setResizeWeight( elements[i].resizeWeight() );
		    final int thisSize = (int)(elements[i].getRelativeSize() * totalSize);
		    q.setDividerLocation( thisSize );
		    p.setRightComponent( q );
		    p = q;
		  }
		else
		  {
		    // We have only one more (this) element, and there is no
		    // divider toward its right side that can be positioned,
		    // it simply gets all remaining space.
		    p.setRightComponent( elements[i].component() );
		  }
	      }
	  }
	invalid = false;
	revalidate();
      }
  }


  /**
   * Describes an element of the MultiSplitPane and its sizing parameters.
   * 
   * @author K Udo Schuermann
   */
  public static class Element
    implements
      Cloneable
  {
    public Element( final Component component )
    {
      this( component,
	    0.5d,
	    0.5d,
	    false );
    }


    public Element( final Component component,
	            final double sizeWeight,
	            final double relativeSize )
    {
      this( component,
	    sizeWeight,
	    relativeSize,
	    true );
    }


    private Element( final Component component,
	             final double sizeWeight,
	             final double relativeSize,
	             final boolean isSized )
    {
      super();
      this.component = component;
      this.resizeWeight = sizeWeight;
      this.relativeSize = relativeSize;
      this.isSized = isSized;
    }


    public Component component()
    {
      return component;
    }


    public double resizeWeight()
    {
      return resizeWeight;
    }


    public void setRelativeSize( final double relativeSize )
    {
      this.relativeSize = relativeSize;
      isSized = true;
    }


    public double getRelativeSize()
    {
      return relativeSize;
    }


    boolean isSized()
    {
      return isSized;
    }


    @Override
    public Element clone()
    {
      return new Element( component,
	                  resizeWeight,
	                  relativeSize );
    }

    private final Component component;
    private final double resizeWeight;
    private double relativeSize;
    private boolean isSized;
  }


  public enum SplitOrientation
  {
    /**
     * Elements are organized left-to-right
     */
    HORIZONTAL(JSplitPane.HORIZONTAL_SPLIT),
    /**
     * Elements are organized top-to-bottom
     */
    VERTICAL(JSplitPane.VERTICAL_SPLIT);

    int getJSplitPaneOrientation()
    {
      return orientation;
    }


    private SplitOrientation( final int orientation )
    {
      this.orientation = orientation;
    }
    private final int orientation;
  }

  private Element[] elements;
  private boolean invalid;
  private Color dividerColor = Color.WHITE;
  private int dividerSize = 12;
  private final SplitOrientation orientation;
  private final boolean continuousLayout;
  private static final long serialVersionUID = -1462225695250741156L;
}
