package com.ringlord.swing;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;

import com.ringlord.ComparableRange;


/**
 * <p>
 * An editor and comparator for a range consisting either of a less-than or
 * greater-than comparison or an inclusive/exclusive two-value range that the
 * user can configure at will. The high-end of the range is optional: omitting
 * it allows entering an open range, but the user can still turn off the
 * high-end range to force an open range. Note: The user can also turn off the
 * entire Range to disable the use of the Range.
 * </p>
 * 
 * <p>
 * The caller is responsible for parsing the contents of the {@link JComponent}
 * objects through which the user edits the low-end and (optional) high-end
 * values. The parsed values can then be range checked against this object using
 * {@link #isMatch(Comparable, Comparable, Comparable)}. For this comparison to
 * be valid, the Range must be prevented from modification (the easiest way is
 * to use {@link #setEnabled(boolean)} with a value of false).
 * </p>
 * 
 * <p>
 * The RangeEditor relies on rules defined by {@link ComparableRange} to
 * highlight certain pointless comparison pairs, such as
 * "less than x and less than y" or
 * "greater than x and greater than or equal to y".
 * </p>
 * 
 * @author K Udo Schuermann
 */
public class RangeEditor
  extends JPanel
{
  /**
   * Construct a RangeEditor using a default visual representation for the
   * comparison operators.
   * 
   * @param firstEnd
   *          The low-end editor for this type. Must not be null.
   * @param secondEnd
   *          The optional high-end editor for this type. Can be null.
   * @see #RangeEditor(JComponent, JComponent, Map)
   */
  public RangeEditor( final JComponent firstEnd,
	              final JComponent secondEnd )
  {
    this( firstEnd,
	  secondEnd,
	  null );
  }


  /**
   * Construct a RangeEditor with a custom visual representation for the
   * comparison functions.
   * 
   * @param firstEnd
   *          The low-end editor for this type. Must not be null.
   * @param secondEnd
   *          The optional high-end editor for this type. Can be null.
   * @param comparisons
   *          An optional {@link Map} (it may be null or empty) to supply
   *          alternate visual representations for any or all of the comparison
   *          operations defined by the {@link ComparableRange.Comparator}. For
   *          a date comparison, for example, you may may prefer the literal
   *          "before" over "&lt;" so your map should contain
   *          <tt>{ComparableRange.Comparator.LT&nbsp;=&gt;&nbsp;"before"}</tt>;
   *          Entries in the map override default based on Unicode mathematical
   *          characters.
   * @see #RangeEditor(JComponent, JComponent)
   */
  public RangeEditor( final JComponent firstEnd,
	              final JComponent secondEnd,
	              final Map<ComparableRange.Comparator,String> comparisons )
  {
    super( new BorderLayout() );
    setOpaque( false );
    JPanel elementRoot = null;

    if( firstEnd == null )
      {
	throw new IllegalArgumentException( "The lowEnd range cannot be null" );
      }

    // Apply custom comparison values (if any) to override defaults
    this.comparisons = new LinkedHashMap<>();
    if( comparisons != null )
      {
	for( final Map.Entry<ComparableRange.Comparator,String> e : comparisons.entrySet() )
	  {
	    this.comparisons.put( e.getKey(),
		                  e.getValue() );
	  }
      }
    for( final ComparableRange.Comparator e : ComparableRange.Comparator.values() )
      {
	if( this.comparisons.get( e ) == null )
	  {
	    this.comparisons.put( e,
		                  e.getVisual() );
	  }
      }

    this.lEnd = new EditPane( firstEnd );
    elementRoot = addLayout( elementRoot,
	                     this.lEnd );
    if( secondEnd != null )
      {
	this.hEnd = new EditPane( secondEnd );
	this.andOrLabel = new JLabel( AND );
	connect( this.lEnd,
	         this.hEnd );

	elementRoot = addLayout( elementRoot,
	                         new JLabel( "   " ) ); // 3 spaces
	elementRoot = addLayout( elementRoot,
	                         andOrLabel );
	elementRoot = addLayout( elementRoot,
	                         new JLabel( "   " ) ); // 3 spaces
	elementRoot = addLayout( elementRoot,
	                         this.hEnd );
      }
    else
      {
	this.hEnd = null;
	this.andOrLabel = null;
      }

    fixLoHi( lEnd,
	     hEnd,
	     false );
  }


  @Override
  public Dimension getMinimumSize()
  {
    return super.getPreferredSize();
  }


  private JPanel addLayout( final JPanel current,
	                    final JComponent addition )
  {
    if( current == null )
      {
	add( BorderLayout.WEST,
	     addition );
	return this;
      }
    else
      {
	final JPanel p = new JPanel( new BorderLayout() );
	p.setOpaque( false );
	p.add( BorderLayout.WEST,
	       addition );
	current.add( BorderLayout.CENTER,
	             p );
	return p;
      }
  }


  @Override
  public void setEnabled( final boolean isEnabled )
  {
    super.setEnabled( isEnabled );
    lEnd.setEnabled( isEnabled );
    if( hEnd != null )
      {
	final boolean lSel = lEnd.getEnabler().isSelected();
	hEnd.setEnabled( lSel && isEnabled );
	andOrLabel.setEnabled( lSel && hEnd.getEnabler().isSelected() && isEnabled );
      }
  }


  /**
   * Alter the value of the primary (low-end) checkbox.
   * 
   * @param isEnabled
   */
  public void setFirstEnabled( final boolean isEnabled )
  {
    lEnd.getEnabler().setSelected( isEnabled );
  }


  /**
   * Obtain the state of the primary (low-end) checkbox.
   * 
   * @return
   */
  public boolean isFirstEnabled()
  {
    return lEnd.getEnabler().isSelected();
  }


  /**
   * Obtain the state of the secondary (high-end) checkbox. If the high-end is
   * not defined (through the Constructor) then this method has no effect.
   * 
   * @param isSecondaryEnabled
   */
  public void setSecondEnabled( final boolean isSecondaryEnabled )
  {
    if( hEnd != null )
      {
	hEnd.getEnabler().setSelected( isSecondaryEnabled );
      }
  }


  /**
   * Alter the state of the secondary (high-end) checkbox. If the high-end is
   * not defined (through the constructor) then this method always returns
   * false.
   * 
   * @return
   */
  public boolean isSecondEnabled()
  {
    return (hEnd == null
	? false
	: hEnd.getEnabler().isSelected());
  }


  /**
   * Set the low-end comparison.
   * 
   * @param comparator
   *          One of the values "LT", "LE", "EQ", "GE", or "GT", irrespective of
   *          the visual representation. Othe values have no effect.
   */
  public void setFirstComparator( final ComparableRange.Comparator comparator )
  {
    lEnd.setComparator( comparator );
  }


  /**
   * Set the high-end comparison. If the high-end is not defined (through the
   * constructor) then this method has no effect.
   * 
   * @param comparator
   *          One of the values "LT", "LE", "EQ", "GE", or "GT", irrespective of
   *          the visual representation. Other values have no effect.
   */
  public void setSecondComparator( final ComparableRange.Comparator comparator )
  {
    if( hEnd != null )
      {
	hEnd.setComparator( comparator );
      }
  }


  /**
   * Obtain the value of the low-end comparison.
   * 
   * @return One of the values "LT", "LE", "EQ", "GE", or "GT"
   */
  public ComparableRange.Comparator getFirstComparison()
  {
    return lEnd.getComparator();
  }


  /**
   * Obtain the value of the high-end comparison.
   * 
   * @return One of the values "LT", "LE", "EQ", "GE", or "GT"; will return ""
   *         (empty String) if the high-end of the range was not defined.
   */
  public ComparableRange.Comparator getSecondComparison()
  {
    return (hEnd == null
	? null
	: hEnd.getComparator());
  }


  /**
   * Obtain the low-end editor.
   * 
   * @return The {@link JComponent} passed to the constructor for editing the
   *         low-end of the range.
   */
  public JComponent getFirstEditor()
  {
    return lEnd.editor;
  }


  /**
   * Obtain the high-end editor.
   * 
   * @return The {@link JComponent} passed to the constructor for editing the
   *         high-end of the range. This will be null if null was passed to the
   *         constructor for this.
   */
  public JComponent getSecondEditor()
  {
    return (hEnd == null
	? null
	: hEnd.editor);
  }


  /**
   * A utility method that connects the various components with listeners to
   * effect dynamic adjustment and (some) validation.
   * 
   * @param lo
   *          The low-end {@link EditPane}
   * @param hi
   *          The high-end {@link EditPane} (if any, it is null if no high-end
   *          editor was defined)
   */
  private void connect( final EditPane lo,
	                final EditPane hi )
  {
    lo.getEnabler().addItemListener( new ItemListener()
    {
      @Override
      public void itemStateChanged( final ItemEvent e )
      {
	if( e.getStateChange() == ItemEvent.SELECTED )
	  {
	    andOrLabel.setEnabled( hi.getEnabler().isSelected() );
	    hi.setEnabled( true );
	  }
	else
	  {
	    andOrLabel.setEnabled( false );
	    hi.setEnabled( false );
	  }
      }
    } );

    lo.getComparatorSelector().addItemListener( new ItemListener()
    {
      @Override
      public void itemStateChanged( final ItemEvent e )
      {
	if( e.getStateChange() == ItemEvent.SELECTED )
	  {
	    if( hi != null )
	      {
		fixLoHi( lo,
		         hi,
		         true );
	      }
	  }
      }
    } );

    hi.getEnabler().addItemListener( new ItemListener()
    {
      @Override
      public void itemStateChanged( final ItemEvent e )
      {
	andOrLabel.setEnabled( hi.getEnabler().isSelected() );
	fixLoHi( lEnd,
	         hEnd,
	         false );
      }
    } );
    hi.getComparatorSelector().addItemListener( new ItemListener()
    {
      @Override
      public void itemStateChanged( final ItemEvent e )
      {
	if( e.getStateChange() == ItemEvent.SELECTED )
	  {
	    fixLoHi( lo,
		     hi,
		     false );
	  }
      }
    } );
  }


  /**
   * One of the comparisons changed, so we need to validate them.
   * 
   * @param lo
   *          The low-end {@link EditPane}
   * @param hi
   *          The high-end {@link EditPane} (if any, it is null if not defined.)
   * @param changedLow
   *          'true' if it was the low-end comparison that changed, rather than
   *          the high-end.
   */
  private void fixLoHi( final EditPane lo,
	                final EditPane hi,
	                final boolean changedLow )
  {
    if( isSecondEnabled() )
      {
	final ComparableRange.Comparator lActual = lo.getComparator();
	final ComparableRange.Comparator hActual = hi.getComparator();

	if( ComparableRange.isIntersectingComparison( lActual,
	                                              hActual ) )
	  {
	    andOrLabel.setText( AND );
	    lo.setValid( true );
	    hi.setValid( true );
	  }
	else if( ComparableRange.isDisjointComparison( lActual,
	                                               hActual ) )
	  {
	    andOrLabel.setText( OR );
	    lo.setValid( true );
	    hi.setValid( true );
	  }
	else
	  {
	    if( changedLow )
	      {
		lo.setValid( false );
		hi.setValid( true );
	      }
	    else
	      {
		lo.setValid( true );
		hi.setValid( false );
	      }
	  }
      }
    else
      {
	lo.setValid( true );
	hi.setValid( true );
      }
  }


  /**
   * A panel that combines a check box, Comparison drop-down, and the
   * user-supplied editor component.
   * 
   * @author K Udo Schuermann
   */
  private class EditPane
    extends JPanel
  {
    EditPane( final JComponent editor )
    {
      super( new BorderLayout() );
      setOpaque( false );
      JPanel elementRoot = null;

      this.editor = editor;
      this.enabled = new JCheckBox();
      this.enabled.setSelected( true );

      final ComparableRange.Comparator[] comparableValues = ComparableRange.Comparator.values();
      final String[] comps = new String[comparableValues.length];
      this.c2v = new HashMap<ComparableRange.Comparator,String>();
      this.v2c = new HashMap<String,ComparableRange.Comparator>();
      int i = 0;
      for( final Map.Entry<ComparableRange.Comparator,String> e : comparisons.entrySet() )
	{
	  final ComparableRange.Comparator k = e.getKey();
	  final String v = e.getValue();
	  comps[i++] = v;
	  c2v.put( k,
		   v );
	  if( v2c.get( v ) != null )
	    {
	      throw new IllegalArgumentException( "Duplicate visual comparison: '" + v + "'" );
	    }
	  v2c.put( v,
		   k );
	}
      this.comparatorSelector = new JComboBox<String>( comps );
      this.comboBackground = comparatorSelector.getBackground();

      elementRoot = addLayout( elementRoot,
	                       enabled );
      elementRoot = addLayout( elementRoot,
	                       comparatorSelector );
      elementRoot = addLayout( elementRoot,
	                       editor );

      enabled.addItemListener( new ItemListener()
      {
	@Override
	public void itemStateChanged( final ItemEvent e )
	{
	  if( e.getStateChange() == ItemEvent.SELECTED )
	    {
	      comparatorSelector.setEnabled( true );
	      editor.setEnabled( true );
	    }
	  else
	    {
	      comparatorSelector.setEnabled( false );
	      editor.setEnabled( false );
	    }
	}
      } );
    }


    private JPanel addLayout( final JPanel current,
	                      final JComponent addition )
    {
      if( current == null )
	{
	  add( BorderLayout.WEST,
	       addition );
	  return this;
	}
      else
	{
	  final JPanel p = new JPanel( new BorderLayout() );
	  p.setOpaque( false );
	  p.add( BorderLayout.WEST,
		 addition );
	  current.add( BorderLayout.CENTER,
		       p );
	  return p;
	}
    }


    @Override
    public void setEnabled( final boolean isEnabled )
    {
      super.setEnabled( isEnabled );
      enabled.setEnabled( isEnabled );
      comparatorSelector.setEnabled( enabled.isSelected() && isEnabled );
      editor.setEnabled( enabled.isSelected() && isEnabled );
    }


    /**
     * Indicates that the Comparison Drop-down selection is not valid when
     * compared to the one in the other editor.
     * 
     * @param isValid
     */
    void setValid( final boolean isValid )
    {
      if( isValid )
	{
	  comparatorSelector.setBackground( comboBackground );
	  comparatorSelector.setToolTipText( null );
	}
      else
	{
	  comparatorSelector.setBackground( Color.red );
	  comparatorSelector.setToolTipText( "<html>"
		                             + "The combination of left and right side<br>"
		                             + "comparisons do not make sense." );
	}
    }


    void setComparator( final ComparableRange.Comparator comparator )
    {
      comparatorSelector.setSelectedItem( c2v.get( comparator ) );
    }


    /**
     * Obtains the selected Comparison value.
     * 
     * @return A value "LT", "LE", "EQ", "GE", or "GT"
     * @see #getComparatorSelector()
     */
    ComparableRange.Comparator getComparator()
    {
      return v2c.get( comparatorSelector.getSelectedItem() );
    }


    /**
     * Obtains the {@link JCheckBox} through which the user can control whether
     * this editor is enabled or not. The low-end editor controls availability
     * of the entire Range; the high-end editor will control only whether the
     * high-end is used.
     * 
     * @return The JCheckBox
     */
    JCheckBox getEnabler()
    {
      return enabled;
    }


    /**
     * Obtains the JComboBox through which the user selects the comparison. Its
     * elements are "LT", "LE", "EQ", "GE", and "GT".
     * 
     * @return The JComboBox
     * @return {@link #getComparator()}
     */
    JComboBox<String> getComparatorSelector()
    {
      return comparatorSelector;
    }

    private final JComponent editor;
    private final JCheckBox enabled;
    private final JComboBox<String> comparatorSelector;
    private final Map<ComparableRange.Comparator,String> c2v;
    private final Map<String,ComparableRange.Comparator> v2c;
    private final Color comboBackground;
    private static final long serialVersionUID = -4491653205185145047L;
  }

  private final JLabel andOrLabel;
  private final EditPane lEnd;
  private final EditPane hEnd;
  private final Map<ComparableRange.Comparator,String> comparisons;
  //
  private static final String AND = " AND ";
  private static final String OR = " OR ";
  private static final long serialVersionUID = 2220583325335183773L;
}
