package com.ringlord.swing;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.AbstractButton;
import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.plaf.basic.BasicButtonUI;


/**
 * A custom JTabbedPane Title Component, icon on the left, title in the center,
 * and optional close button on the right.
 * 
 * @author K Udo Schuermann
 */
public class TabTitle
  extends JPanel
{
  /**
   * Adds a TabTitle to the indicated {@link JTabbedPane}.
   * 
   * @param tabs
   *          The JTabbedPane to receive the TabTitle.
   * @param index
   *          The index where the TabTitle is to be added. Use a negative number
   *          to added it as the last tab.
   * @param title
   *          The title text for the TabTitle. This value may be null if no
   *          title is desired (in which case you ought to have at least an
   *          icon).
   * @param icon
   *          The icon for the TabTitle. This value may be null if no icon is
   *          desired (in which case you ought to have at least a title.
   * @param component
   *          The {@link Component} to be shown when this TabTitle is selected
   *          (the component is the "content" of the tab, i.e. the potentially
   *          large amount of information to be made visible when this tab is
   *          selected).
   * @param addCloseButton
   *          Indicate whether to add a close button to the TabTitle. Note that
   *          pressing the close button
   *          <em>automatically removes the entire tab from the JTabbedPane</em>
   *          If you want to be notified when the user has clicked the button,
   *          add an {@link ActionListener} to the
   *          {@link TabTitle#getCloseButton()}. If you want to control whether
   *          the tab is actually closed, then you must remove the TabTitle's
   *          ActionListener from the button and add your own.
   */
  public static TabTitle addTab( final JTabbedPane tabs,
	                         final int index,
	                         final String title,
	                         final Icon icon,
	                         final Component component,
	                         final boolean addCloseButton )
  {
    int tabIndex = (index < 0
	? tabs.getTabCount()
	: index);

    tabs.insertTab( title,
	            icon,
	            component,
	            null,
	            tabIndex );
    final TabTitle tabTitle = new TabTitle( tabs,
	                                    addCloseButton );
    if( addCloseButton )
      {
	tabTitle.getCloseButton().addActionListener( new ActionListener()
	{
	  public void actionPerformed( final ActionEvent e )
	  {
	    final int index = tabs.indexOfTabComponent( tabTitle );
	    if( index > -1 )
	      {
		tabs.setTabComponentAt( index,
		                        null );
	      }
	    tabs.removeTabAt( index );
	  }
	} );
      }
    tabs.setTabComponentAt( tabIndex,
	                    tabTitle );
    return tabTitle;
  }


  public TabTitle( final JTabbedPane tabs,
	           final boolean addCloseButton )
  {
    super( new FlowLayout( FlowLayout.LEFT,
	                   0,
	                   0 ) );
    setOpaque( false );
    final TabTitle tabTitle = this;
    this.title = new JLabel()
    {
      public Icon getIcon()
      {
	final int index = tabs.indexOfTabComponent( tabTitle );
	if( index > -1 )
	  {
	    return tabs.getIconAt( index );
	  }
	return null;
      }


      public void setIcon( final Icon icon )
      {
	final int index = tabs.indexOfTabComponent( tabTitle );
	if( index > -1 )
	  {
	    tabs.setIconAt( index,
		            icon );
	    revalidate();
	  }
      }


      public String getText()
      {
	final int index = tabs.indexOfTabComponent( tabTitle );
	if( index > -1 )
	  {
	    return tabs.getTitleAt( index );
	  }
	return null;
      }


      public void setText( final String text )
      {
	final int index = tabs.indexOfTabComponent( tabTitle );
	if( index > -1 )
	  {
	    tabs.setTitleAt( index,
		             text );
	    revalidate();
	  }
      }
      private static final long serialVersionUID = -9053887467207771208L;
    };
    this.title.setOpaque( false );
    title.setBorder( BorderFactory.createEmptyBorder( 0,
	                                              0,
	                                              0,
	                                              5 ) );
    add( title );
    title.addMouseListener( new MouseAdapter()
    {
      public void mouseClicked( final MouseEvent e )
      {
	final Component clicked = e.getComponent();
	for( int i = tabs.getTabCount() - 1; i >= 0; i-- )
	  {
	    final Component c = tabs.getTabComponentAt( i );
	    if( (c == clicked) || ((c instanceof TabTitle) && ((TabTitle)c).title == clicked) )
	      {
		tabs.setSelectedIndex( i );
		break;
	      }
	  }
      }
    } );

    final JPanel p = new JPanel( new BorderLayout() );
    p.setOpaque( false );
    if( addCloseButton )
      {
	p.add( this.close = new CloseButton() );
      }
    else
      {
	this.close = null;
      }
    p.setBorder( BorderFactory.createEmptyBorder( 1,
	                                          0,
	                                          0,
	                                          0 ) );
    add( p );

    setBorder( BorderFactory.createEmptyBorder( 2,
	                                        0,
	                                        0,
	                                        0 ) );
  }


  @Override
  public void setToolTipText( final String text )
  {
    super.setToolTipText( text );
    title.setToolTipText( text );
  }


  public void setTitle( final String title )
  {
    this.title.setText( title );
  }


  public String getTitle()
  {
    return this.title.getText();
  }


  public void setIcon( final Icon icon )
  {
    this.title.setIcon( icon );
  }


  public Icon getIcon()
  {
    return title.getIcon();
  }


  /**
   * Obtain a reference to the Close Button on the Tab Component, provided that
   * the TabComponent was created with that option. Two obvious uses for this
   * method are:
   * 
   * <ol>
   * <li>Add an {@link ActionListener} to hear when the user clicks this button
   * (which causes the tab to be removed from its JTabbedPane),
   * <li>Add a tooltip text to the button, such as "Close this tab".
   * </ol>
   * 
   * @return The close button, or null if no close button was added.
   */
  public AbstractButton getCloseButton()
  {
    return close;
  }


  private class CloseButton
    extends JButton
  {
    CloseButton()
    {
      setPreferredSize( new Dimension( 14,
	                               14 ) );

      setUI( new BasicButtonUI() );
      setContentAreaFilled( false );
      setFocusable( false );
      setBorder( BorderFactory.createEtchedBorder() );
      setBorderPainted( false );

      addMouseListener( closeButtonMouseListener );
      setRolloverEnabled( true );
      setOpaque( false );
    }


    public void updateUI()
    {
      // do nothing, we don't want this UI to be messed with
    }


    protected void paintComponent( final Graphics g )
    {
      super.paintComponent( g );

      final Graphics2D g2 = (Graphics2D)g.create();
      g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
	                   RenderingHints.VALUE_ANTIALIAS_ON );

      if( getModel().isPressed() && mouseInside )
	{
	  // when pressed, the image is shifted slightly for 3D illusion
	  g2.translate( 1,
		        1 );
	}

      final int wide = getWidth();
      final int high = getHeight();
      g2.setStroke( stroke2 );
      if( mouseInside ) // roll-over
	{
	  g2.setColor( getBackground() );
	  g2.fillRect( 0,
		       0,
		       wide,
		       high );
	  g2.setColor( Color.red );
	}
      else
	{
	  g2.setColor( Color.black );
	}

      final int w = wide - d - 1;
      final int h = high - d - 1;
      g2.drawLine( d,
	           d,
	           w,
	           h );
      g2.drawLine( w,
	           d,
	           d,
	           h );
      g2.dispose();
    }

    private final static int d = 4;
    private static final long serialVersionUID = -4808819655693963652L;
  }
  private final MouseListener closeButtonMouseListener = new MouseAdapter()
  {
    public void mouseEntered( final MouseEvent e )
    {
      final Component c = e.getComponent();
      if( c instanceof AbstractButton )
	{
	  ((AbstractButton)c).setBorderPainted( true );
	  mouseInside = true;
	}
    }


    public void mouseExited( final MouseEvent e )
    {
      final Component c = e.getComponent();
      if( c instanceof AbstractButton )
	{
	  ((AbstractButton)c).setBorderPainted( false );
	  mouseInside = false;
	}
    }
  };

  private boolean mouseInside;
  private final JLabel title;
  private final JButton close;
  private final static BasicStroke stroke2 = new BasicStroke( 1 );
  private static final long serialVersionUID = 607256873278952373L;
}
