package com.ringlord.swing;

import java.awt.Color;
import java.awt.Event;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.lang.ref.WeakReference;
import java.util.HashSet;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.InputMap;
import javax.swing.JTextPane;
import javax.swing.KeyStroke;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;
import javax.swing.text.StyledEditorKit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoManager;


/**
 * A simple text editor with support for bold, italic, and underlined attributes
 * that can be applied using the traditional ctrl-b, ctrl-i, and ctrl-u keys;
 * Undo/redo functionality is bound to the ctrl-z/shift-ctrl-z key sequences.
 * Also supported are {@linkplain #Link external content references} (links).
 * 
 * @author K Udo Schuermann
 */
public class TextEditor
  extends JTextPane
  implements
    ChangeListener
{
  /**
   * Constructs a Text editor.
   */
  public TextEditor()
  {
    super();
    setEditable( true );
    setCaretPosition( 0 );

    this.undoAction = new UndoAction();
    this.redoAction = new RedoAction();

    initStyles();
    bindKeys();
  }


  /**
   * Insert text, replacing the current selection (if any).
   * 
   * @param text
   */
  public void insertText( final String text )
  {
    replaceSelection( "" );
    final int n = getCaretPosition();
    replaceSelection( text );
    setCaretPosition( n + text.length() );
  }


  /**
   * Insert a {@link Link}, replacing the current selection, if any.
   * 
   * @param link
   */
  public void insertLink( final Link link )
  {
    links.add( new WeakReference<Link>( link ) );

    link.addChangeListener( this );

    link.setFont( getFont() );
    replaceSelection( "" );
    final int n = getCaretPosition();
    insertComponent( link );
    setCaretPosition( n + 1 );

    revalidate();
    repaint();
  }


  /**
   * Set the number of pixels to be used for spacing between paragraphs.
   * 
   * @param below
   */
  public void setParagraphSpacing( final int below )
  {
    final SimpleAttributeSet sas = new SimpleAttributeSet();
    StyleConstants.setSpaceBelow( sas,
	                          below );
    StyleConstants.setFontFamily( sas,
	                          "Serif" );
    ((StyledDocument)getDocument()).setParagraphAttributes( 0,
	                                                    Integer.MAX_VALUE,
	                                                    sas,
	                                                    true );
  }


  private void initStyles()
  {
    final StyledDocument doc = getStyledDocument();

    // The defaultStyle is the basis from which we derive all others
    final Style defaultStyle = StyleContext.getDefaultStyleContext().getStyle( StyleContext.DEFAULT_STYLE );
    StyleConstants.setFontFamily( defaultStyle,
	                          "Monospaced" );
    StyleConstants.setFontSize( defaultStyle,
	                        11 );

    doc.addStyle( "plain",
	          defaultStyle );

    // The bold style, as its name so cleverly suggests, is bold
    final Style bold = doc.addStyle( "bold",
	                             defaultStyle );
    StyleConstants.setBold( bold,
	                    true );

    // Continuing the tradition of obviousness, the italic style is ... italic
    final Style italics = doc.addStyle( "italic",
	                                defaultStyle );
    StyleConstants.setItalic( italics,
	                      true );

    // Underline
    final Style underline = doc.addStyle( "underline",
	                                  defaultStyle );
    StyleConstants.setUnderline( underline,
	                         true );

    // Links to internal and external resources are underlined and blue
    final Style link = doc.addStyle( "link",
	                             defaultStyle );
    StyleConstants.setUnderline( link,
	                         true );
    StyleConstants.setForeground( link,
	                          Color.blue );

    doc.addUndoableEditListener( new UndoEditListener() );
  }


  private void bindKeys()
  {
    final InputMap inputMap = getInputMap();
    // final int kbdMask = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()

    // Bold, Italic, Underline styles
    bind( inputMap,
	  KeyStroke.getKeyStroke( KeyEvent.VK_B,
	                          Event.CTRL_MASK ),
	  "bold",
	  new StyledEditorKit.BoldAction() );
    bind( inputMap,
	  KeyStroke.getKeyStroke( KeyEvent.VK_I,
	                          Event.CTRL_MASK ),
	  "italic",
	  new StyledEditorKit.ItalicAction() );
    bind( inputMap,
	  KeyStroke.getKeyStroke( KeyEvent.VK_U,
	                          Event.CTRL_MASK ),
	  "underline",
	  new StyledEditorKit.UnderlineAction() );

    // Undo/Redo capability
    bind( inputMap,
	  KeyStroke.getKeyStroke( KeyEvent.VK_Z,
	                          Event.CTRL_MASK ),
	  "undo",
	  undoAction );
    bind( inputMap,
	  KeyStroke.getKeyStroke( KeyEvent.VK_Z,
	                          Event.SHIFT_MASK | Event.CTRL_MASK ),
	  "redo",
	  redoAction );
  }


  private static void bind( final InputMap inputMap,
	                    final KeyStroke key,
	                    final String styleName,
	                    final Action action )
  {
    action.putValue( Action.NAME,
	             styleName );
    inputMap.put( key,
	          action );
  }


  // This one listens for edits that can be undone.
  private class UndoEditListener
    implements
      UndoableEditListener
  {
    @Override
    public void undoableEditHappened( UndoableEditEvent e )
    {
      // Remember the edit and update the menus.
      undoManager.addEdit( e.getEdit() );
      undoAction.updateUndoState();
      redoAction.updateRedoState();
    }
  }


  private class UndoAction
    extends AbstractAction
  {
    public UndoAction()
    {
      super( "Undo" );
      setEnabled( false );
    }


    @Override
    public void actionPerformed( ActionEvent e )
    {
      try
	{
	  undoManager.undo();
	}
      catch( CannotUndoException ex )
	{
	  System.out.println( "Unable to undo: " + ex );
	  ex.printStackTrace();
	}
      updateUndoState();
      redoAction.updateRedoState();
    }


    protected void updateUndoState()
    {
      if( undoManager.canUndo() )
	{
	  setEnabled( true );
	  putValue( Action.NAME,
		    undoManager.getUndoPresentationName() );
	}
      else
	{
	  setEnabled( false );
	  putValue( Action.NAME,
		    "Undo" );
	}
    }
    private static final long serialVersionUID = -7057529118533986529L;
  }


  private class RedoAction
    extends AbstractAction
  {
    public RedoAction()
    {
      super( "Redo" );
      setEnabled( false );
    }


    @Override
    public void actionPerformed( ActionEvent e )
    {
      try
	{
	  undoManager.redo();
	}
      catch( CannotRedoException ex )
	{
	  System.out.println( "Unable to redo: " + ex );
	  ex.printStackTrace();
	}
      updateRedoState();
      undoAction.updateUndoState();
    }


    protected void updateRedoState()
    {
      if( undoManager.canRedo() )
	{
	  setEnabled( true );
	  putValue( Action.NAME,
		    undoManager.getRedoPresentationName() );
	}
      else
	{
	  setEnabled( false );
	  putValue( Action.NAME,
		    "Redo" );
	}
    }
    private static final long serialVersionUID = 5325069684976655965L;
  }


  @Override
  public void stateChanged( ChangeEvent e )
  {
    repaint();
  }

  private final UndoAction undoAction;
  private final RedoAction redoAction;
  private final UndoManager undoManager = new UndoManager();
  private final Set<WeakReference<Link>> links = new HashSet<>();
  //
  private static final long serialVersionUID = 3039744033536689637L;
}
