package com.ringlord.swing;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import com.ringlord.convert.Value;
import com.ringlord.convert.Value.Unit;


/**
 * A numeric entry field paired with a dropdown to define the units of the
 * numeric entry. Appending units to the numeric entry changes the unit
 * dropdown's value. Changing the unit dropdown manually converts the value to
 * the new units.
 * 
 * @author K Udo Schuermann
 */
public class UnitField
  extends JPanel
{
  private static final long serialVersionUID = -1369143921749457916L;


  public static void main( final String[] args )
  {
    final javax.swing.JFrame f = new javax.swing.JFrame( "Unit Testing" );
    f.setDefaultCloseOperation( javax.swing.WindowConstants.EXIT_ON_CLOSE );
    final javax.swing.JPanel p = new javax.swing.JPanel( new java.awt.GridLayout( 0,
	                                                                          1 ) );
    p.add( new UnitField( Value.Type.MASS ) );
    p.add( new UnitField( Value.Type.LENGTH ) );
    p.add( new UnitField( Value.Type.TEMPERATURE ) );
    p.add( new UnitField( Value.Type.LUMINOSITY ) );
    f.setContentPane( p );
    f.pack();
    f.setVisible( true );
  }


  public UnitField( final Value.Type type )
  {
    super( new BorderLayout() );

    this.type = type;

    this.text = new JTextField( 4 );
    this.text.setFont( font );
    this.text.getDocument().addDocumentListener( entryHandler );

    this.units = new JComboBox<Value.Unit>( TYPE_CHOICES.get( type ) );
    this.units.setFont( font );
    this.units.setRenderer( new UnitsRenderer() );
    this.units.addItemListener( conversionListener );
    this.units.setEnabled( false );

    this.oldUnits = (Value.Unit)this.units.getSelectedItem();

    add( BorderLayout.WEST,
	 new JLabel( type + ": " ) );
    add( BorderLayout.CENTER,
	 text );
    add( BorderLayout.EAST,
	 units );
  }


  public UnitField setValue( final double value )
  {
    text.setText( String.valueOf( value ) );
    return this;
  }


  public UnitField setValue( final double value,
	                     final Value.Unit unit )
  {
    if( unit.type() != type )
      {
	throw new IllegalArgumentException( "Expected a unit of type " + type + ", not " + unit.type() );
      }
    units.removeItemListener( conversionListener );
    text.setText( String.valueOf( this.value = value ) );
    units.setSelectedItem( unit );
    units.addItemListener( conversionListener );
    return this;
  }


  public UnitField setValue( final String valuePlusUnit )
  {
    return setValue( Value.parse( valuePlusUnit,
	                          (Value.Type)units.getSelectedItem() ) );
  }


  public UnitField setValue( final Value value )
  {
    return setValue( value.doubleValue(),
	             value.unit() );
  }


  public Value getValue()
  {
    return new Value( getDouble(),
	              getUnit() );
  }


  public Value.Unit getUnit()
  {
    return (Value.Unit)units.getSelectedItem();
  }


  public double getDouble()
  {
    return value;
  }


  private class EntryHandler
    implements
      DocumentListener
  {
    public void changedUpdate( DocumentEvent e )
    {
      process();
    }


    public void insertUpdate( DocumentEvent e )
    {
      process();
    }


    public void removeUpdate( DocumentEvent e )
    {
      process();
    }


    private void process()
    {
      if( text.isFocusOwner() )
	{
	  final String s = text.getText();
	  try
	    {
	      value = Double.parseDouble( s );
	      //
	      text.setForeground( normalColor );
	      units.setEnabled( true );
	    }
	  catch( Exception x )
	    {
	      try
		{
		  Value v = Value.parse( s,
			                 oldUnits.type() );
		  value = v.doubleValue();
		  oldUnits = v.unit();
		  units.setSelectedItem( oldUnits );
		  //
		  text.setForeground( normalColor );
		  units.setEnabled( true );
		}
	      catch( Exception x2 )
		{
		  text.setForeground( errorColor );
		  units.setEnabled( false );
		}
	    }
	}
    }
    private final Color normalColor = Color.BLACK;
    private final Color errorColor = Color.RED;
  }


  private class ConversionListener
    implements
      ItemListener
  {
    @Override
    public void itemStateChanged( ItemEvent e )
    {
      if( units.isFocusOwner() )
	{
	  try
	    {
	      final Value.Unit toUnits = (Value.Unit)units.getSelectedItem();
	      if( toUnits != oldUnits )
		{
		  final double v = oldUnits.convert( value,
			                             toUnits );
		  System.err.println( "Convert " +
			              value +
			              " " +
			              oldUnits.suffix() +
			              " to " +
			              v +
			              " " +
			              toUnits.suffix() );
		  value = v;
		  text.setText( String.valueOf( v ) );
		  oldUnits = toUnits;
		}
	    }
	  catch( Exception x )
	    {
	      x.printStackTrace();
	    }
	}
    }
  }


  private class UnitsRenderer
    extends JLabel
    implements
      ListCellRenderer<Value.Unit>
  {
    private static final long serialVersionUID = -9144168476692220531L;


    UnitsRenderer()
    {
      super();
      setOpaque( true );
    }


    @Override
    public Component getListCellRendererComponent( final JList<? extends Unit> list,
	                                           final Unit value,
	                                           final int index,
	                                           final boolean isSelected,
	                                           final boolean hasFocus )
    {
      setText( value.suffix() );
      if( isSelected )
	{
	  setForeground( list.getSelectionForeground() );
	  setBackground( list.getSelectionBackground() );
	}
      else
	{
	  setForeground( list.getForeground() );
	  setBackground( list.getBackground() );
	}
      return this;
    }
  }

  private static final Map<Value.Type,Value.Unit[]> TYPE_CHOICES = new HashMap<Value.Type,Value.Unit[]>();
  static
    {
      // Dynamically build the map TYPE_CHOICES map
      final Map<Value.Type,List<Value.Unit>> m = new HashMap<Value.Type,List<Value.Unit>>();
      for( Value.Unit u : Value.Unit.values() )
	{
	  final Value.Type t = u.type();
	  List<Value.Unit> units = m.get( t );
	  if( units == null )
	    {
	      units = new ArrayList<Value.Unit>();
	      m.put( t,
		     units );
	    }
	  units.add( u );
	}
      for( Map.Entry<Value.Type,List<Value.Unit>> e : m.entrySet() )
	{
	  final List<Value.Unit> v = e.getValue();
	  Value.Unit[] u = new Value.Unit[v.size()];
	  v.toArray( u );
	  TYPE_CHOICES.put( e.getKey(),
		            u );
	}
    }

  private double value;
  private Value.Unit oldUnits;
  private final Value.Type type;
  private final JTextField text;
  private final JComboBox<Value.Unit> units;
  private final EntryHandler entryHandler = new EntryHandler();
  private final ConversionListener conversionListener = new ConversionListener();
  private static final Font font = new Font( "SansSerif",
	                                     Font.PLAIN,
	                                     11 );
}
